#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""
BDD acceptance tests management.

Exports

*  ``bdd_acceptance_command`` function
"""

from .bdd_group import bdd_acceptance_command  # noqa: F401



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import copy
import pathlib
import subprocess

import pytest

from build_harness.commands import main
from build_harness.commands.unit_tests import (
    CommandOptionsTestTool,
    CoverageTestToolError,
    ExitState,
    UnitTestError,
    _run_tests,
)
from tests.ci.support.click_runner import click_runner  # noqa: F401

MOCK_VENV_PATH = pathlib.Path("/some/test_unit_test/path")


class TestRunTests:
    MOCK_OPTIONS: CommandOptionsTestTool = {
        "output_path": pathlib.Path("dist/path"),
        "source_path": pathlib.Path("source/path"),
        "test_path": pathlib.Path("test/path"),
        "venv_path": pathlib.Path("venv/path"),
        "pass_zero_tests": False,
        "report_enabled": {
            "junitxml": False,
            "term-missing": True,
            "html": False,
            "xml": False,
        },
        "report_dirs": {
            "junitxml": None,
            "term-missing": None,
            "html": None,
            "xml": None,
        },
    }

    def test_clean(self, mocker):
        mock_result = mocker.MagicMock()
        mock_result.returncode = 0
        mocker.patch(
            "build_harness.commands.unit_tests.run_command",
            return_value=mock_result,
        )

        _run_tests(self.MOCK_OPTIONS, None, False)

    def test_failed_raises(self, mocker):
        mock_result = mocker.MagicMock()
        mock_result.returncode = 1
        mocker.patch(
            "build_harness.commands.unit_tests.run_command",
            return_value=mock_result,
        )

        with pytest.raises(UnitTestError, match=r"^unit tests failed"):
            _run_tests(self.MOCK_OPTIONS, None, False)

    def test_above_threshold(self, mocker):
        mock_captured_report = """some/file.py                     5      0   100%
--------------------------------------------------------
TOTAL                                  170     44    91%
"""
        mock_result = mocker.MagicMock()
        mock_result.returncode = 0
        mock_result.stdout = mock_captured_report
        mocker.patch(
            "build_harness.commands.unit_tests.run_command",
            return_value=mock_result,
        )

        _run_tests(self.MOCK_OPTIONS, 90, False)

    def test_equal_threshold(self, mocker):
        mock_captured_report = """some/file.py                     5      0   100%
--------------------------------------------------------
TOTAL                                  170     44    90%
"""
        mock_result = mocker.MagicMock()
        mock_result.returncode = 0
        mock_result.stdout = mock_captured_report
        mocker.patch(
            "build_harness.commands.unit_tests.run_command",
            return_value=mock_result,
        )

        _run_tests(self.MOCK_OPTIONS, 90, False)

    def test_below_threshold(self, mocker):
        mock_captured_report = """some/file.py                     5      0   100%
--------------------------------------------------------
TOTAL                                  170     44    89%
"""
        mock_result = mocker.MagicMock()
        mock_result.returncode = 0
        mock_result.stdout = mock_captured_report
        mocker.patch(
            "build_harness.commands.unit_tests.run_command",
            return_value=mock_result,
        )

        with pytest.raises(
            CoverageTestToolError, match=r"^Coverage test failed"
        ):
            _run_tests(self.MOCK_OPTIONS, 90, False)

    def test_100_percent(self, mocker) -> None:
        mock_captured_report = """some/file.py                     5      0   100%
--------------------------------------------------------
TOTAL                                  170     44    100%
"""
        mock_result = mocker.MagicMock()
        mock_result.returncode = 0
        mock_result.stdout = mock_captured_report
        mocker.patch(
            "build_harness.commands.unit_tests.run_command",
            return_value=mock_result,
        )

        _run_tests(self.MOCK_OPTIONS, 90, False)

    def test_0_percent(self, mocker) -> None:
        mock_captured_report = """some/file.py                     5      0   100%
--------------------------------------------------------
TOTAL                                  170     44    0%
"""
        mock_result = mocker.MagicMock()
        mock_result.returncode = 0
        mock_result.stdout = mock_captured_report
        mocker.patch(
            "build_harness.commands.unit_tests.run_command",
            return_value=mock_result,
        )

        _run_tests(self.MOCK_OPTIONS, 0, False)

    def test_no_report_raises(self, mocker) -> None:
        mock_result = mocker.MagicMock()
        mock_result.returncode = 0
        mock_result.stdout = None
        mocker.patch(
            "build_harness.commands.unit_tests.run_command",
            return_value=mock_result,
        )

        with pytest.raises(
            CoverageTestToolError, match=r"^no coverage report captured"
        ):
            _run_tests(self.MOCK_OPTIONS, 0, False)

    def test_no_tests_raises(self, mocker):
        """Zero tests raises an exception."""
        mock_options = copy.deepcopy(self.MOCK_OPTIONS)
        mock_options["pass_zero_tests"] = False

        mock_result = mocker.MagicMock()
        mock_result.returncode = 5
        mocker.patch(
            "build_harness.commands.unit_tests.run_command",
            return_value=mock_result,
        )

        with pytest.raises(UnitTestError, match=r"^unit tests failed"):
            _run_tests(mock_options, None, False)

    def test_no_tests_fail_disabled(self, mocker):
        """Disabling failure for no tests does not raise."""
        mock_options = copy.deepcopy(self.MOCK_OPTIONS)
        mock_options["pass_zero_tests"] = True

        mock_result = mocker.MagicMock()
        mock_result.returncode = 5
        mocker.patch(
            "build_harness.commands.unit_tests.run_command",
            return_value=mock_result,
        )

        _run_tests(mock_options, None, False)


class TestUnitTest:
    def test_pass(self, click_runner, mocker, mock_sysargv):
        mock_pytest_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_pytest_result.returncode = 0
        mock_pytest_result.stdout = b""
        mock_run = mocker.patch(
            "build_harness.commands.unit_tests.run_command",
            side_effect=[mock_pytest_result],
        )
        mocker.patch(
            "build_harness.commands.unit_tests.acquire_source_dir",
            return_value="the/source/path",
        )
        result = click_runner.invoke(main, ["unit-test"])

        assert result.exit_code == 0
        mock_run.assert_has_calls(
            [
                mocker.call(
                    ["/some/conftest/path/pytest", "tests"],
                    capture_output=True,
                ),
            ]
        )

    def test_fail(self, click_runner, mocker, mock_sysargv):
        mock_pytest_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_pytest_result.returncode = 1
        mock_pytest_result.stdout = b""
        mock_run = mocker.patch(
            "build_harness.commands.unit_tests.run_command",
            side_effect=[mock_pytest_result],
        )
        mocker.patch(
            "build_harness.commands.unit_tests.acquire_source_dir",
            return_value="the/source/path",
        )
        result = click_runner.invoke(main, ["unit-test"])

        assert result.exit_code == ExitState.TESTS_FAILED.value
        mock_run.assert_has_calls(
            [
                mocker.call(
                    ["/some/conftest/path/pytest", "tests"],
                    capture_output=True,
                ),
            ]
        )


@pytest.fixture()
def mock_sourcedir(mocker):
    mocker.patch(
        "build_harness.commands.unit_tests.acquire_source_dir",
        return_value="source/path",
    )


@pytest.fixture()
def mock_run(mock_sysargv, mock_sourcedir, mocker):
    mock_pytest_result = mocker.create_autospec(subprocess.CompletedProcess)
    mock_pytest_result.returncode = 0
    mock_pytest_result.stdout = b""
    this_run = mocker.patch(
        "build_harness.commands.unit_tests.run_command",
        side_effect=[mock_pytest_result],
    )

    return this_run


class TestCoverageReports:
    def test_coverage_console(self, click_runner, mocker, mock_run):
        result = click_runner.invoke(
            main,
            [
                "unit-test",
                "--coverage-console",
            ],
        )

        assert result.exit_code == 0
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        "/some/conftest/path/pytest",
                        "--cov=source/path",
                        "--cov-report",
                        "term-missing",
                        "tests",
                    ],
                    capture_output=True,
                ),
            ]
        )

    def test_coverage_html_auto(self, click_runner, mocker, mock_run):
        result = click_runner.invoke(main, ["unit-test", "--coverage-html"])

        assert result.exit_code == 0
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        "/some/conftest/path/pytest",
                        "--cov=source/path",
                        "--cov-report",
                        "html:dist/coverage_report",
                        "tests",
                    ],
                    capture_output=True,
                ),
            ]
        )

    def test_coverage_html_explicit(self, click_runner, mocker, mock_run):
        result = click_runner.invoke(
            main,
            [
                "unit-test",
                "--coverage-html",
                "--coverage-html-file",
                "some/html/path",
            ],
        )

        assert result.exit_code == 0
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        "/some/conftest/path/pytest",
                        "--cov=source/path",
                        "--cov-report",
                        "html:some/html/path",
                        "tests",
                    ],
                    capture_output=True,
                ),
            ]
        )

    def test_coverage_xml_auto(self, click_runner, mocker, mock_run):
        result = click_runner.invoke(main, ["unit-test", "--coverage-xml"])

        assert result.exit_code == 0
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        "/some/conftest/path/pytest",
                        "--cov=source/path",
                        "--cov-report",
                        "xml:dist/coverage_report.xml",
                        "tests",
                    ],
                    capture_output=True,
                ),
            ]
        )

    def test_coverage_xml_explicit(self, click_runner, mocker, mock_run):
        result = click_runner.invoke(
            main,
            [
                "unit-test",
                "--coverage-xml",
                "--coverage-xml-file",
                "some/xml/file",
            ],
        )

        assert result.exit_code == 0
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        "/some/conftest/path/pytest",
                        "--cov=source/path",
                        "--cov-report",
                        "xml:some/xml/file",
                        "tests",
                    ],
                    capture_output=True,
                ),
            ]
        )

    def test_multiple_coverage(self, click_runner, mocker, mock_run):
        result = click_runner.invoke(
            main,
            [
                "unit-test",
                "--coverage-console",
                "--coverage-html",
                "--coverage-xml",
            ],
        )

        assert result.exit_code == 0
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        "/some/conftest/path/pytest",
                        "--cov=source/path",
                        "--cov-report",
                        "html:dist/coverage_report",
                        "--cov-report",
                        "term-missing",
                        "--cov-report",
                        "xml:dist/coverage_report.xml",
                        "tests",
                    ],
                    capture_output=True,
                ),
            ]
        )


class TestJunitReport:
    def test_junitxml_auto(self, click_runner, mocker, mock_run):
        result = click_runner.invoke(main, ["unit-test", "--junitxml"])

        assert result.exit_code == 0
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        "/some/conftest/path/pytest",
                        "--cov=source/path",
                        "--junitxml=dist/junit_report.xml",
                        "tests",
                    ],
                    capture_output=True,
                ),
            ]
        )

    def test_junitxml_explicit(self, click_runner, mocker, mock_run):
        result = click_runner.invoke(
            main, ["unit-test", "--junitxml-file", "some/junit/file"]
        )

        assert result.exit_code == 0
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        "/some/conftest/path/pytest",
                        "--cov=source/path",
                        "--junitxml=some/junit/file",
                        "tests",
                    ],
                    capture_output=True,
                ),
            ]
        )


class TestCoverageCheck:
    mock_captured_report = """some/file.py                     5      0   100%
    --------------------------------------------------------
    TOTAL                                  170     44    91%
    """

    def test_pass(self, click_runner, mocker, mock_sysargv, mock_sourcedir):
        mock_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_result.returncode = 0
        mock_result.stdout = self.mock_captured_report
        mock_run = mocker.patch(
            "build_harness.commands.unit_tests.run_command",
            return_value=mock_result,
        )
        result = click_runner.invoke(main, ["unit-test", "--check", "91"])

        assert result.exit_code == 0
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        "/some/conftest/path/pytest",
                        "--cov=source/path",
                        "--cov-report",
                        "term-missing",
                        "tests",
                    ],
                    capture_output=True,
                ),
            ]
        )

    def test_fail(self, click_runner, mocker, mock_sysargv, mock_sourcedir):
        mock_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_result.returncode = 0
        mock_result.stdout = self.mock_captured_report
        mock_run = mocker.patch(
            "build_harness.commands.unit_tests.run_command",
            return_value=mock_result,
        )
        result = click_runner.invoke(main, ["unit-test", "--check", "92"])

        assert result.exit_code == ExitState.TEST_COVERAGE_FAILED.value
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        "/some/conftest/path/pytest",
                        "--cov=source/path",
                        "--cov-report",
                        "term-missing",
                        "tests",
                    ],
                    capture_output=True,
                ),
            ]
        )

    def test_pass_zero_tests(self, click_runner, mocker, mock_sysargv):
        mock_pytest_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_pytest_result.returncode = 5
        mock_pytest_result.stdout = ""
        mock_run = mocker.patch(
            "build_harness.commands.unit_tests.run_command",
            side_effect=[mock_pytest_result],
        )
        mocker.patch(
            "build_harness.commands.unit_tests.acquire_source_dir",
            return_value="the/source/path",
        )
        result = click_runner.invoke(main, ["unit-test", "--pass-zero-tests"])

        assert result.exit_code == 0
        mock_run.assert_has_calls(
            [
                mocker.call(
                    ["/some/conftest/path/pytest", "tests"],
                    capture_output=True,
                ),
            ]
        )



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#
import abc
import pathlib

import git


class TagBase(metaclass=abc.ABCMeta):
    def __init__(self, tag: str):
        self.tag = tag

    def _commit_file(self, file_path: pathlib.Path, repo: git.Repo) -> None:
        filename = str(file_path)
        open(filename, "wb").close()
        repo.index.add([filename])
        repo.index.commit(filename)

    @abc.abstractmethod
    def __call__(self, repo: git.Repo, repo_dir: pathlib.Path) -> None:
        return


class NoTags(TagBase):
    def __init__(self):
        super().__init__("no_tag")

    def __call__(self, repo: git.Repo, repo_dir: pathlib.Path) -> None:
        filename = repo_dir / "c1"
        self._commit_file(filename, repo)


class TagOnHead(TagBase):
    def __init__(self, tag: str):
        super().__init__(tag)

    def __call__(self, repo: git.Repo, repo_dir: pathlib.Path) -> None:
        filename = repo_dir / "c1"
        self._commit_file(filename, repo)
        repo.create_tag(self.tag, message="default branch tag")


class TagOnDefaultHeadOnFeature(TagBase):
    def __init__(self, tag: str):
        super().__init__(tag)

    def __call__(self, repo: git.Repo, repo_dir: pathlib.Path) -> None:
        self._commit_file(repo_dir / "c1", repo)
        repo.create_tag(self.tag, message="default branch tag")
        self._commit_file(repo_dir / "c2", repo)
        nb = repo.create_head("feature")
        repo.head.reference = nb
        repo.head.reset(index=True, working_tree=True)

        assert repo.active_branch.name == "feature"

        self._commit_file(repo_dir / "fc1", repo)


class FeatureDryrunTagOnHead(TagBase):
    def __init__(self, tag: str):
        super().__init__(tag)

    def __call__(self, repo: git.Repo, repo_dir: pathlib.Path) -> None:
        self._commit_file(repo_dir / "c1", repo)
        repo.create_tag(self.tag, message="default branch tag")
        self._commit_file(repo_dir / "c2", repo)
        nb = repo.create_head("feature")
        repo.head.reference = nb
        repo.head.reset(index=True, working_tree=True)

        assert repo.active_branch.name == "feature"

        self._commit_file(repo_dir / "fc1", repo)
        repo.create_tag("3.1+dryrun", message="feature branch dryrun")
        self._commit_file(repo_dir / "fc2", repo)
        repo.create_tag("3.1+dryrun2", message="feature branch dryrun 2")


class FeatureDryrunTags(TagBase):
    def __init__(self, tag: str):
        super().__init__(tag)

    def __call__(self, repo: git.Repo, repo_dir: pathlib.Path) -> None:
        self._commit_file(repo_dir / "c1", repo)
        repo.create_tag(self.tag, message="default branch tag")
        self._commit_file(repo_dir / "c2", repo)
        nb = repo.create_head("feature")
        nb.commit = repo.head.commit
        repo.head.reference = nb
        repo.head.reset(index=True, working_tree=True)

        assert repo.active_branch.name == "feature"

        self._commit_file(repo_dir / "fc1", repo)
        repo.create_tag("3.1+dryrun", message="feature branch dryrun")
        self._commit_file(repo_dir / "fc2", repo)
        repo.create_tag("3.1+dryrun2", message="feature branch dryrun 2")
        self._commit_file(repo_dir / "fc3", repo)



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import copy
import pathlib

import pytest

from build_harness.project import (
    DependencyValidationError,
    ProjectDependencyAnalysis,
    PyprojecttomlError,
    _acquire_pyprojecttoml_data,
    acquire_project_dependencies,
    acquire_source_dir,
)

MOCK_PROJECT_DIR = pathlib.Path("some/test_project/path")


class TestAcquirePyprojecttomlData:
    def test_flit(self, mocker):
        mocker.patch(
            "build_harness.project.pathlib.Path.is_file", return_value=True
        )
        mock_open = mocker.mock_open(
            read_data="""
[tool.flit.metadata]
module = "some_source_dir"
"""
        )
        mocker.patch("build_harness.project.pathlib.Path.open", mock_open)

        result = _acquire_pyprojecttoml_data(MOCK_PROJECT_DIR)

        assert result == {
            "tool": {"flit": {"metadata": {"module": "some_source_dir"}}}
        }

    def test_no_pyprojecttoml(self, mocker):
        mocker.patch(
            "build_harness.project.pathlib.Path.is_file", return_value=False
        )

        with pytest.raises(
            PyprojecttomlError, match="^Missing pyproject.toml file"
        ):
            _acquire_pyprojecttoml_data(MOCK_PROJECT_DIR)


class TestFlitProject:
    def test_flit(self, mocker):
        this_mock_data = {
            "build-system": {
                "requires": [
                    "flit >=2, <4",
                ],
            },
            "tool": {
                "flit": {
                    "metadata": {"module": "sample_project"},
                },
            },
        }
        mocker.patch(
            "build_harness.project._acquire_pyprojecttoml_data",
            return_value=this_mock_data,
        )

        result = acquire_source_dir(MOCK_PROJECT_DIR)

        assert result == "sample_project"

    def test_flit_core_old_style(self, mocker):
        this_mock_data = {
            "build-system": {
                "requires": [
                    "flit_core >=3, <4",
                ],
            },
            "tool": {
                "flit": {
                    "metadata": {"module": "sample_project"},
                },
            },
        }
        mocker.patch(
            "build_harness.project._acquire_pyprojecttoml_data",
            return_value=this_mock_data,
        )

        result = acquire_source_dir(MOCK_PROJECT_DIR)

        assert result == "sample_project"

    def test_flit_core_new_style(self, mocker):
        this_mock_data = {
            "build-system": {
                "requires": [
                    "flit_core >=3, <4",
                ],
            },
            "project": {"name": "sample_project"},
        }
        mocker.patch(
            "build_harness.project._acquire_pyprojecttoml_data",
            return_value=this_mock_data,
        )

        result = acquire_source_dir(MOCK_PROJECT_DIR)

        assert result == "sample_project"

    def test_missing_flit_project_name(self, mocker):
        this_mock_data = {
            "build-system": {
                "requires": [
                    "flit_core >=3, <4",
                ],
            },
        }
        mocker.patch(
            "build_harness.project._acquire_pyprojecttoml_data",
            return_value=this_mock_data,
        )
        with pytest.raises(
            PyprojecttomlError,
            match=r"^Malformed flit pyproject\.toml missing project name",
        ):
            acquire_source_dir(MOCK_PROJECT_DIR)

    def test_bad_build_system_raises(self, mocker):
        this_mock_data = {
            "build-system": {
                "requires": [
                    "invalid-system",
                ],
            },
        }
        mocker.patch(
            "build_harness.project._acquire_pyprojecttoml_data",
            return_value=this_mock_data,
        )

        with pytest.raises(
            PyprojecttomlError, match="^Invalid build system specified"
        ):
            acquire_source_dir(MOCK_PROJECT_DIR)

    def test_missing_build_system_raises(self, mocker):
        this_mock_data = {
            "build-system": {},
        }
        mocker.patch(
            "build_harness.project._acquire_pyprojecttoml_data",
            return_value=this_mock_data,
        )

        with pytest.raises(
            PyprojecttomlError,
            match="^Invalid pyproject.toml, missing build system specification",
        ):
            acquire_source_dir(MOCK_PROJECT_DIR)


class TestAcquireProjectDependencies:
    def test_flit_old_style_dependencies(self, mocker):
        mocker.patch(
            "build_harness.project.pathlib.Path.is_file", return_value=True
        )
        mock_open = mocker.mock_open(
            read_data="""
[tool.flit.metadata]
requires = [
    "requests >=2.2, <3.0",
    "toml == 1.0.0",
]

[tool.flit.metadata.requires-extra]
dev = [
    "pre_commit == 2.7.1",
]
doc = [
    "sphinx == 3.2.1",
]
test = [
    "pytest == 6.1.1",
]
"""
        )
        mocker.patch("build_harness.project.pathlib.Path.open", mock_open)

        result = acquire_project_dependencies(MOCK_PROJECT_DIR)

        expected_data = {
            "runtime": [
                "requests >=2.2, <3.0",
                "toml == 1.0.0",
            ],
            "dev": ["pre_commit == 2.7.1"],
            "doc": ["sphinx == 3.2.1"],
            "test": ["pytest == 6.1.1"],
        }
        assert result == expected_data

    def test_flit_new_style_dependencies(self, mocker):
        mocker.patch(
            "build_harness.project.pathlib.Path.is_file", return_value=True
        )
        mock_open = mocker.mock_open(
            read_data="""
[project]
requires = [
    "requests >=2.2, <3.0",
    "toml == 1.0.0",
]

[project.optional-dependencies]
dev = [
    "pre_commit == 2.7.1",
]
doc = [
    "sphinx == 3.2.1",
]
test = [
    "pytest == 6.1.1",
]
"""
        )
        mocker.patch("build_harness.project.pathlib.Path.open", mock_open)

        result = acquire_project_dependencies(MOCK_PROJECT_DIR)

        expected_data = {
            "runtime": [
                "requests >=2.2, <3.0",
                "toml == 1.0.0",
            ],
            "dev": ["pre_commit == 2.7.1"],
            "doc": ["sphinx == 3.2.1"],
            "test": ["pytest == 6.1.1"],
        }
        assert result == expected_data

    def test_missing_requires(self, mocker):
        mocker.patch(
            "build_harness.project.pathlib.Path.is_file", return_value=True
        )
        mock_open = mocker.mock_open(
            read_data="""
[tool.flit.metadata]

[tool.flit.metadata.requires-extra]
dev = [
    "pre_commit == 2.7.1",
]
doc = [
    "sphinx == 3.2.1",
]
test = [
    "pytest == 6.1.1",
]
"""
        )
        mocker.patch("build_harness.project.pathlib.Path.open", mock_open)

        result = acquire_project_dependencies(MOCK_PROJECT_DIR)

        expected_data = {
            "runtime": [],
            "dev": ["pre_commit == 2.7.1"],
            "doc": ["sphinx == 3.2.1"],
            "test": ["pytest == 6.1.1"],
        }
        assert result == expected_data

    def test_missing_dev(self, mocker):
        mocker.patch(
            "build_harness.project.pathlib.Path.is_file", return_value=True
        )
        mock_open = mocker.mock_open(
            read_data="""
[tool.flit.metadata]
requires = [
    "requests >=2.2, <3.0",
    "toml == 1.0.0",
]

[tool.flit.metadata.requires-extra]
doc = [
    "sphinx == 3.2.1",
]
test = [
    "pytest == 6.1.1",
]
"""
        )
        mocker.patch("build_harness.project.pathlib.Path.open", mock_open)

        result = acquire_project_dependencies(MOCK_PROJECT_DIR)

        expected_data = {
            "runtime": [
                "requests >=2.2, <3.0",
                "toml == 1.0.0",
            ],
            "dev": [],
            "doc": ["sphinx == 3.2.1"],
            "test": ["pytest == 6.1.1"],
        }
        assert result == expected_data

    def test_missing_doc(self, mocker):
        mocker.patch(
            "build_harness.project.pathlib.Path.is_file", return_value=True
        )
        mock_open = mocker.mock_open(
            read_data="""
[tool.flit.metadata]
requires = [
    "requests >=2.2, <3.0",
    "toml == 1.0.0",
]

[tool.flit.metadata.requires-extra]
dev = [
    "pre_commit == 2.7.1",
]
test = [
    "pytest == 6.1.1",
]
"""
        )
        mocker.patch("build_harness.project.pathlib.Path.open", mock_open)

        result = acquire_project_dependencies(MOCK_PROJECT_DIR)

        expected_data = {
            "runtime": [
                "requests >=2.2, <3.0",
                "toml == 1.0.0",
            ],
            "dev": ["pre_commit == 2.7.1"],
            "doc": [],
            "test": ["pytest == 6.1.1"],
        }
        assert result == expected_data

    def test_missing_test(self, mocker):
        mocker.patch(
            "build_harness.project.pathlib.Path.is_file", return_value=True
        )
        mock_open = mocker.mock_open(
            read_data="""
[tool.flit.metadata]
requires = [
    "requests >=2.2, <3.0",
    "toml == 1.0.0",
]

[tool.flit.metadata.requires-extra]
dev = [
    "pre_commit == 2.7.1",
]
doc = [
    "sphinx == 3.2.1",
]
"""
        )
        mocker.patch("build_harness.project.pathlib.Path.open", mock_open)

        result = acquire_project_dependencies(MOCK_PROJECT_DIR)

        expected_data = {
            "runtime": [
                "requests >=2.2, <3.0",
                "toml == 1.0.0",
            ],
            "dev": ["pre_commit == 2.7.1"],
            "doc": ["sphinx == 3.2.1"],
            "test": [],
        }
        assert result == expected_data

    def test_invalid_project(self, mocker):
        mocker.patch(
            "build_harness.project.pathlib.Path.is_file", return_value=True
        )
        mock_open = mocker.mock_open(
            read_data="""
[not.a.project]
"""
        )
        mocker.patch("build_harness.project.pathlib.Path.open", mock_open)
        with pytest.raises(
            PyprojecttomlError,
            match=r"^Invalid pyproject\.toml project specification",
        ):
            acquire_project_dependencies(MOCK_PROJECT_DIR)


class TestProjectDependencyAnalysis:
    # https://www.python.org/dev/peps/pep-0440/#version-specifiers
    MOCK_DEPENDENCIES = {
        "runtime": [
            "compatible1 ~= 3.1.4",
            "compatible2 ~= 3.1",
            "version_matching == 3.1.4",
            "version_exclusion != 3.1.4",
            "inclusive_ordered1 <=3.1.4",
            "inclusive_ordered2 >=3.1.4",
            "exclusive_ordered1 <3.1.4",
            "exclusive_ordered2 >3.1.4",
            "arbitrary_equality === custom4rel",
            "runtime_exclusive == 3.1.4",
            "multiple_specifiers >=2.2, <3.0",
            "empty_specifier",
        ],
        "dev": ["dev_exclusive == 3.1.4"],
        "doc": ["doc_exclusive == 3.1.4"],
        "test": ["test_exclusive == 3.1.4"],
    }

    def test_all(self):
        under_test = ProjectDependencyAnalysis(
            copy.deepcopy(self.MOCK_DEPENDENCIES), "all"
        )

        assert under_test.packages == [
            "compatible1",
            "compatible2",
            "version_matching",
            "version_exclusion",
            "inclusive_ordered1",
            "inclusive_ordered2",
            "exclusive_ordered1",
            "exclusive_ordered2",
            "arbitrary_equality",
            "runtime_exclusive",
            "multiple_specifiers",
            "empty_specifier",
            "dev_exclusive",
            "doc_exclusive",
            "test_exclusive",
        ]

        assert under_test.valid_release("runtime_exclusive", "3.1.4")
        assert under_test.valid_release("dev_exclusive", "3.1.4")
        assert under_test.valid_release("doc_exclusive", "3.1.4")
        assert under_test.valid_release("test_exclusive", "3.1.4")

    def test_runtime(self):
        under_test = ProjectDependencyAnalysis(
            copy.deepcopy(self.MOCK_DEPENDENCIES), "runtime"
        )

        assert under_test.packages == [
            "compatible1",
            "compatible2",
            "version_matching",
            "version_exclusion",
            "inclusive_ordered1",
            "inclusive_ordered2",
            "exclusive_ordered1",
            "exclusive_ordered2",
            "arbitrary_equality",
            "runtime_exclusive",
            "multiple_specifiers",
            "empty_specifier",
        ]

    def test_dev(self):
        under_test = ProjectDependencyAnalysis(
            copy.deepcopy(self.MOCK_DEPENDENCIES), "dev"
        )

        assert under_test.packages == [
            "dev_exclusive",
        ]

    def test_doc(self):
        under_test = ProjectDependencyAnalysis(
            copy.deepcopy(self.MOCK_DEPENDENCIES), "doc"
        )

        assert under_test.packages == [
            "doc_exclusive",
        ]

    def test_test(self):
        under_test = ProjectDependencyAnalysis(
            copy.deepcopy(self.MOCK_DEPENDENCIES), "test"
        )

        assert under_test.packages == [
            "test_exclusive",
        ]

    def test_compatible1(self):
        under_test = ProjectDependencyAnalysis(
            copy.deepcopy(self.MOCK_DEPENDENCIES), "all"
        )

        assert not under_test.valid_release("compatible1", "4.0")
        assert not under_test.valid_release("compatible1", "3.1.3")
        assert not under_test.valid_release("compatible1", "3.2")
        assert under_test.valid_release("compatible1", "3.1.4")
        assert under_test.valid_release("compatible1", "3.1.5")

    def test_compatible2(self):
        under_test = ProjectDependencyAnalysis(
            copy.deepcopy(self.MOCK_DEPENDENCIES), "all"
        )

        assert not under_test.valid_release("compatible2", "4.0")
        assert not under_test.valid_release("compatible2", "3.0")
        assert under_test.valid_release("compatible2", "3.1.3")
        assert under_test.valid_release("compatible2", "3.2")

    def test_version_matching(self):
        under_test = ProjectDependencyAnalysis(
            copy.deepcopy(self.MOCK_DEPENDENCIES), "all"
        )

        assert under_test.valid_release("version_matching", "3.1.4")
        assert not under_test.valid_release("version_matching", "3.1.3")
        assert not under_test.valid_release("version_matching", "3.1.5")

    def test_version_exclusion(self):
        under_test = ProjectDependencyAnalysis(
            copy.deepcopy(self.MOCK_DEPENDENCIES), "all"
        )

        assert not under_test.valid_release("version_exclusion", "3.1.4")
        assert under_test.valid_release("version_exclusion", "3.1.3")
        assert under_test.valid_release("version_exclusion", "3.1.5")

    def test_inclusive_ordered1(self):
        under_test = ProjectDependencyAnalysis(
            copy.deepcopy(self.MOCK_DEPENDENCIES), "all"
        )

        assert under_test.valid_release("inclusive_ordered1", "3.1.4")
        assert under_test.valid_release("inclusive_ordered1", "3.1.3")
        assert not under_test.valid_release("inclusive_ordered1", "3.1.5")

    def test_inclusive_ordered2(self):
        under_test = ProjectDependencyAnalysis(
            copy.deepcopy(self.MOCK_DEPENDENCIES), "all"
        )

        assert under_test.valid_release("inclusive_ordered2", "3.1.4")
        assert not under_test.valid_release("inclusive_ordered2", "3.1.3")
        assert under_test.valid_release("inclusive_ordered2", "3.1.5")

    def test_exclusive_ordered1(self):
        under_test = ProjectDependencyAnalysis(
            copy.deepcopy(self.MOCK_DEPENDENCIES), "all"
        )

        assert not under_test.valid_release("exclusive_ordered1", "3.1.4")
        assert under_test.valid_release("exclusive_ordered1", "3.1.3")
        assert not under_test.valid_release("exclusive_ordered1", "3.1.5")

    def test_exclusive_ordered2(self):
        under_test = ProjectDependencyAnalysis(
            copy.deepcopy(self.MOCK_DEPENDENCIES), "all"
        )

        assert not under_test.valid_release("exclusive_ordered2", "3.1.4")
        assert not under_test.valid_release("exclusive_ordered2", "3.1.3")
        assert under_test.valid_release("exclusive_ordered2", "3.1.5")

    @pytest.mark.skip("arbitrary equality not supported")
    def test_arbitrary_equality(self):
        under_test = ProjectDependencyAnalysis(
            copy.deepcopy(self.MOCK_DEPENDENCIES), "all"
        )

        assert under_test.valid_release("arbitrary_equality", "custom4rel")
        assert not under_test.valid_release("arbitrary_equality", "3.1.3")

    def test_multiple_specifiers(self):
        mock_dependencies = {
            "runtime": [
                "multiple_specifiers >=2.2, <3.0",
            ],
            "dev": list(),
            "doc": list(),
            "test": list(),
        }
        under_test = ProjectDependencyAnalysis(mock_dependencies, "all")

        assert under_test.valid_release("multiple_specifiers", "2.2")
        assert not under_test.valid_release("multiple_specifiers", "2.1")
        assert not under_test.valid_release("multiple_specifiers", "3.0")

    def test_empty_specifier(self):
        """any release id is valid for an empty specifier"""
        under_test = ProjectDependencyAnalysis(
            copy.deepcopy(self.MOCK_DEPENDENCIES), "all"
        )

        assert under_test.valid_release("empty_specifier", "3.1.4")
        assert under_test.valid_release("empty_specifier", "1")
        assert under_test.valid_release("empty_specifier", "22")

    def test_missing_package_raises(self):
        under_test = ProjectDependencyAnalysis(
            copy.deepcopy(self.MOCK_DEPENDENCIES), "all"
        )
        with pytest.raises(
            DependencyValidationError,
            match=r"^Package not found in dependencies",
        ):
            under_test.valid_release("bad_package", "3.1.4")

    def test_runtime_exclusive(self):
        under_test = ProjectDependencyAnalysis(
            copy.deepcopy(self.MOCK_DEPENDENCIES), "runtime"
        )
        assert under_test.valid_release("runtime_exclusive", "3.1.4")
        assert not under_test.valid_release("runtime_exclusive", "4")
        with pytest.raises(
            DependencyValidationError,
            match=r"^Package not found in dependencies",
        ):
            under_test.valid_release("dev_exclusive", "3.1.4")

    def test_dev_exclusive(self):
        under_test = ProjectDependencyAnalysis(
            copy.deepcopy(self.MOCK_DEPENDENCIES), "dev"
        )
        assert under_test.valid_release("dev_exclusive", "3.1.4")
        with pytest.raises(
            DependencyValidationError,
            match=r"^Package not found in dependencies",
        ):
            under_test.valid_release("runtime_exclusive", "3.1.4")

    def test_doc_exclusive(self):
        under_test = ProjectDependencyAnalysis(
            copy.deepcopy(self.MOCK_DEPENDENCIES), "doc"
        )
        assert under_test.valid_release("doc_exclusive", "3.1.4")
        with pytest.raises(
            DependencyValidationError,
            match=r"^Package not found in dependencies",
        ):
            under_test.valid_release("dev_exclusive", "3.1.4")

    def test_test_exclusive(self):
        under_test = ProjectDependencyAnalysis(
            copy.deepcopy(self.MOCK_DEPENDENCIES), "test"
        )
        assert under_test.valid_release("test_exclusive", "3.1.4")
        with pytest.raises(
            DependencyValidationError,
            match=r"^Package not found in dependencies",
        ):
            under_test.valid_release("dev_exclusive", "3.1.4")



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import pathlib

from behave import given, then

from build_harness._utility import run_command


def _create_mock_pyproject(file_path: pathlib.Path):
    with file_path.open(mode="w") as f:
        f.write(
            """
[tool.flit.metadata]
module = "some_module"

requires = [
    "requests",
]
"""
        )


def create_bad_pyproject(file_path: pathlib.Path):
    with file_path.open(mode="w") as f:
        f.write(
            """
[tool.flit.metadata]
module = "some_module"

requires = [
    "hopefullyreallyimprobablepackage",
]
"""
        )


def dependencies_post_step_actions(context, result, mock_project):
    if result.returncode == 0:
        project_pip_path = mock_project / ".venv" / "bin" / "pip"
        assert project_pip_path.parent.is_dir()
        assert project_pip_path.is_file()
        # only generate this data if the command succeeds.
        result = run_command(
            [str(project_pip_path), "list"],
            capture_output=True,
            text=True,
            universal_newlines=True,
        )
        context.venv_packages = result.stdout
    else:
        context.venv_err = result.stderr


@given("the virtual environment does not exist in the project directory")
def step_impl(context):
    context.existing_project_venv = False
    context.create_pyproject = _create_mock_pyproject
    context.given_arguments = ["install"]
    context.given_expected_package = "requests"
    context.post_step_actions = dependencies_post_step_actions


@given(
    "the project includes configuration that will cause the installation to fail"
)
def step_impl(context):
    context.existing_project_venv = False
    context.create_pyproject = create_bad_pyproject
    context.given_arguments = ["install"]
    context.post_step_actions = dependencies_post_step_actions


@given("the virtual environment already exists in the project directory")
def step_impl(context):
    context.existing_project_venv = True
    context.create_pyproject = _create_mock_pyproject
    context.given_arguments = ["install"]
    context.given_expected_package = "requests"
    context.post_step_actions = dependencies_post_step_actions


@then("installation error is reported to the console")
def step_impl(context):
    print(context.run_result.stdout)
    print(context.run_result.stderr)
    assert "Dependency installation failed" in context.run_result.stderr


@then("the virtual environment is created")
def step_impl(context):
    assert any([str(x).endswith(".venv") for x in context.project_dir_list])


@then(
    "the virtual environment is populated with necessary project dependencies"
)
def step_impl(context):
    print(context.given_expected_package)
    print(context.venv_packages)
    assert context.given_expected_package in context.venv_packages


@given("a virtual environment with the installed project dependencies")
def step_impl(context):
    context.existing_project_venv = True
    context.given_expected_package = "requests"
    context.post_step_actions = dependencies_post_step_actions


def create_requests_version(file_path: pathlib.Path, release_id: str):
    with file_path.open(mode="w") as f:
        f.write(
            """
[tool.flit.metadata]
module = "some_module"

requires = [
    "requests=={0}",
]
""".format(
                release_id
            )
        )


def dependency_change_actions(data: dict):
    project_dir = data["working_dir"]
    run_command(
        [str(project_dir / ".venv/bin/pip"), "install", "requests==2.17.2"],
        capture_output=True,
        text=True,
        universal_newlines=True,
    )


@given(
    "the project includes a dependency change in pyproject.toml that is not "
    "installed in the virtual environment"
)
def step_impl(context):
    context.create_pyproject = lambda x: create_requests_version(x, "2.24.0")
    context.pre_step_actions = dependency_change_actions
    context.post_step_actions = dependencies_post_step_actions


@given("the --check argument is added to the install command run")
def step_impl(context):
    context.given_arguments = ["install", "--check"]


@then("the dependency installation error is reported to the console")
def step_impl(context):
    print(context.run_result.stdout)
    print(context.run_result.stderr)
    assert (
        "Installed packages do not comply with declared project dependencies"
        in context.run_result.stderr
    )



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import contextlib
import dataclasses
import io
import os
import pathlib
import sys
import tempfile
import typing

from build_harness._utility import run_command


@dataclasses.dataclass
class ProjectContext:
    repo_dir: pathlib.Path
    temp_path: pathlib.Path
    template_dir: pathlib.Path
    venv_bin: pathlib.Path
    venv_dir: pathlib.Path


@dataclasses.dataclass
class StdIo:
    out: io.IOBase
    err: io.IOBase


@dataclasses.dataclass()
class ProjectData:
    captured_io: StdIo
    file_contents: typing.Dict[str, typing.Optional[str]]
    project_name: str


@contextlib.contextmanager
def capture_stdout() -> typing.Generator[StdIo, None, None]:
    original_stdio = StdIo(out=sys.stdout, err=sys.stderr)
    try:
        captured_stdio = StdIo(out=io.StringIO(), err=io.StringIO())
        sys.stdout = captured_stdio.out
        sys.stderr = captured_stdio.err

        yield captured_stdio
    finally:
        sys.stdout = original_stdio.out
        sys.stderr = original_stdio.err


def _setup_context(this_dir: pathlib.Path) -> ProjectContext:
    repo_dir = pathlib.Path(__file__).parent / "../../.."
    template_dir = repo_dir / "build_harness/templates"

    temp_path = pathlib.Path(this_dir)
    venv_dir = temp_path / ".venv"
    venv_bin = temp_path / ".venv/bin"

    run_command(["python3", "-m", "venv", str(venv_dir)])
    # trick the install to not download build-harness by pre-installing the
    # local build-harness package
    run_command([str(venv_bin / "pip"), "install", ".[dev,doc,test]"])

    this_context = ProjectContext(
        temp_path=temp_path,
        repo_dir=repo_dir,
        template_dir=template_dir,
        venv_dir=venv_dir,
        venv_bin=venv_bin,
    )

    return this_context


@contextlib.contextmanager
def build_cookiecutter_context(
    working_dir: typing.Optional[pathlib.Path] = None,
) -> typing.Generator[ProjectContext, None, None]:
    if not working_dir:
        with tempfile.TemporaryDirectory() as this_dir:
            this_context = _setup_context(this_dir)
            yield this_context
    else:
        # use the specified path as the working directory
        working_context = _setup_context(working_dir)
        yield working_context


def make_project_venv(
    project_name: str, temp_path: pathlib.Path
) -> pathlib.Path:
    project_dir = temp_path / project_name
    project_venv = project_dir / ".venv"
    project_venv_bin = project_venv / "bin"

    # install this build_harness to acquire correct dependencies
    os.makedirs(str(project_dir), exist_ok=True)
    run_command(["python3", "-m", "venv", str(project_venv)])
    run_command([str(project_venv_bin / "pip"), "install", "."])

    return project_dir


@contextlib.contextmanager
def run_cookiecutter(
    run_context: dict, project_name: str
) -> typing.Generator[typing.Tuple[ProjectContext, StdIo], None, None]:
    with capture_stdout() as captured_stdio:
        # Need to do the import here to ensure that cookiecutter stdio is captured
        # correctly.
        from cookiecutter.main import cookiecutter

        with build_cookiecutter_context() as this_context:
            make_project_venv(project_name, this_context.temp_path)

            cookiecutter(
                str(this_context.template_dir),
                extra_context=run_context,
                no_input=True,
                replay=False,
                overwrite_if_exists=True,
                output_dir=this_context.temp_path,
            )

            yield this_context, captured_stdio



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""Package publish subcommand implementation."""

import copy
import io
import logging
import operator
import pathlib
import sys
import typing

import click

from build_harness._utility import (
    CommandArgs,
    command_path,
    report_console_error,
    run_command,
)

from ._declarations import PublishOptions
from .state import CommandState, ExitState

log = logging.getLogger(__name__)

TWINE_PUBLISH_CMD: CommandArgs = ["twine", "upload", "--non-interactive"]

PYPI_TEST_URL = "https://test.pypi.org/legacy/"


class PublishingError(Exception):
    """Problem occurred during publishing."""


def _obfuscate_password(password: str) -> str:
    """
    Obfuscate a password for debug logging.

    Args:
        password: Password text to be obfuscated.

    Returns:
        Last four characters of password, if possible. Completely obfuscated
        otherwise.
    """
    obfuscated_password: str = (
        "**{0}".format(password[-4:]) if len(password) > 4 else "*****"
    )

    return obfuscated_password


def _publish_packages(
    venv_path: pathlib.Path,
    publish_option: PublishOptions,
    password: typing.Optional[str],
    user: typing.Optional[str],
) -> None:
    """
    Publish sdist, wheel packages using twine.

    On dry run no publish occurs, but a summary of packages found that would
    have been published is printed to console.

    Args:
        venv_path: Path to Python virtual environment.
        publish_option: Publish option selected.
        password: Remote repository password.
        user: Remote repository username

    Raises:
        FileNotFoundError: if password file does not exist.
        PublishingError: If package publish exits non-zero.
    """
    this_command = copy.deepcopy(TWINE_PUBLISH_CMD)
    this_command[0] = command_path(venv_path, this_command)
    if publish_option == PublishOptions.test:
        this_command += ["--repository-url", PYPI_TEST_URL]
    this_command += ["dist/*.whl", "dist/*.tar.gz"]

    environment_variables = {}
    if user:
        log.debug("twine username, {0}".format(user))
        environment_variables["TWINE_USERNAME"] = user
    else:
        log.warning(
            "twine username not provided. assuming twine can figure it out..."
        )
    if password:
        obfuscated_password = _obfuscate_password(password)
        log.debug("twine password, {0}".format(obfuscated_password))
        environment_variables["TWINE_PASSWORD"] = password
    else:
        log.warning(
            "twine password not provided. assuming twine can figure it out..."
        )

    if publish_option in [PublishOptions.yes, PublishOptions.test]:
        log.debug("twine command run, {0}".format(str(this_command)))
        result = run_command(
            this_command,
            env=environment_variables,
            # suppress command logging in the run to protect accidental
            # exposure of credentials
            suppress_command_logging=True,
            suppress_argument_logging=True,
        )

        if any(x != 0 for x in [result.returncode]):
            raise PublishingError("twine failed during package publishing.")
    elif publish_option == PublishOptions.dryrun:
        message = "Dry run, {0}".format(this_command)
        log.warning(message)
        click.echo(message)
    elif publish_option == PublishOptions.no:
        message = "Publish disabled"
        log.info(message)
        click.echo(message)
    else:
        # it shouldn't be possible to get here, unless PublishOptions is
        # changed in future in a way that the other conditionals don't cover.
        raise PublishingError("Unknown publish option")


@click.command()
@click.pass_context
@click.option(
    "--password",
    default=None,
    help="PEP-503 server login password",
    type=str,
)
@click.option(
    "--password-file",
    default=None,
    help="Path to file containing PEP-503 server login password",
    type=click.File(mode="r"),
)
@click.option(
    "--publish",
    default=PublishOptions.yes.name,
    help="""Control whether or not to publish package.

    yes - publish to pypi.org.
    no - do not publish or dryrun.
    dryrun - log the twine command to be run on publish without uploading
             anything.
    test - publish to pypi.org test server.
""",
    show_default=True,
    type=click.Choice(list(map(operator.attrgetter("name"), PublishOptions))),
)
@click.option(
    "--user",
    default=None,
    help="PEP-503 server login user name",
    type=str,
)
def publish(
    ctx: click.Context,
    password: typing.Optional[str],
    password_file: typing.Optional[io.FileIO],
    publish: str,
    user: typing.Optional[str],
) -> None:
    """Publish project artifacts."""
    try:
        ctx.ensure_object(dict)
        command_state: CommandState = ctx.obj["command_state"]

        this_password = None
        if password and password_file:
            raise PublishingError(
                "Cannot specify both `--password` and `--password-file` "
                "options."
            )
        elif password:
            log.debug(
                "password passed as CLI argument, {0}".format(
                    _obfuscate_password(password)
                )
            )
            this_password = password
        elif password_file:
            log.debug(
                "password specified in password file, {0!r}".format(
                    password_file.name
                )
            )
            this_password = typing.cast(str, password_file.read())
            if not this_password:
                raise PublishingError("Empty password file specified, {0}")
            log.debug(
                "password recovered from file, {0}".format(
                    _obfuscate_password(this_password)
                )
            )

        _publish_packages(
            command_state.venv_path,
            PublishOptions[publish],
            this_password,
            user,
        )
    except PublishingError as e:
        message = str(e)
        report_console_error(message)
        sys.exit(ExitState.PUBLISHING_FAILED.value)
    except Exception:
        message = "Unexpected error. Check log for details."
        log.exception(message)
        report_console_error(message)
        sys.exit(ExitState.UNKNOWN_ERROR.value)



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import pathlib
import subprocess

import pytest

import build_harness.commands.build_harness_group
from build_harness.commands import ExitState, main
from build_harness.commands.code_style import (
    BlackCheckError,
    FormattingError,
    IsortCheckError,
    _apply_formatting,
    _check_formatting,
)
from tests.ci.support.click_runner import click_runner  # noqa: F401

MOCK_VENV_PATH = pathlib.Path("/some/test_code_style/path")


class TestApplyFormatting:
    def test_clean(self, mocker):
        mock_result = mocker.MagicMock()
        mock_result.returncode = 0
        mocker.patch(
            "build_harness.commands.code_style.run_command",
            return_value=mock_result,
        )

        _apply_formatting(MOCK_VENV_PATH)

    def test_black_error(self, mocker):
        mock_results = [mocker.MagicMock(), mocker.MagicMock()]
        mock_results[0].returncode = 0
        mock_results[1].returncode = 1
        mocker.patch(
            "build_harness.commands.code_style.run_command",
            side_effect=mock_results,
        )

        with pytest.raises(
            FormattingError,
            match=r"^isort and/or black failed during formatting.",
        ):
            _apply_formatting(MOCK_VENV_PATH)

    def test_isort_error(self, mocker):
        mock_results = [mocker.MagicMock(), mocker.MagicMock()]
        mock_results[0].returncode = 1
        mock_results[1].returncode = 0
        mocker.patch(
            "build_harness.commands.code_style.run_command",
            side_effect=mock_results,
        )

        with pytest.raises(
            FormattingError,
            match=r"^isort and/or black failed during formatting.",
        ):
            _apply_formatting(MOCK_VENV_PATH)


class TestCheckFormatting:
    def test_clean(self, mocker):
        mock_result = mocker.MagicMock()
        mock_result.returncode = 0
        mocker.patch(
            "build_harness.commands.code_style.run_command",
            return_value=mock_result,
        )

        _check_formatting(MOCK_VENV_PATH)

    def test_black_failure(self, mocker):
        mock_results = [mocker.MagicMock(), mocker.MagicMock()]
        mock_results[0].returncode = 0
        mock_results[1].returncode = 1
        mocker.patch(
            "build_harness.commands.code_style.run_command",
            side_effect=mock_results,
        )

        with pytest.raises(BlackCheckError, match=r"^black check failed"):
            _check_formatting(MOCK_VENV_PATH)

    def test_isort_failure(self, mocker):
        mock_results = [mocker.MagicMock(), mocker.MagicMock()]
        mock_results[0].returncode = 1
        mock_results[1].returncode = 0
        mocker.patch(
            "build_harness.commands.code_style.run_command",
            side_effect=mock_results,
        )

        with pytest.raises(IsortCheckError, match=r"^isort check failed"):
            _check_formatting(MOCK_VENV_PATH)


class TestFormatting:
    def test_clean(self, click_runner, mocker):
        mock_isort_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_isort_result.returncode = 0
        mock_black_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_black_result.returncode = 0
        mock_run = mocker.patch(
            "build_harness.commands.code_style.run_command",
            side_effect=[mock_isort_result, mock_black_result],
        )
        build_harness.commands.build_harness_group.sys.argv = [
            "/some/test_code_style_formatting_clean/path/build-harness"
        ]
        result = click_runner.invoke(main, ["formatting"])

        assert result.exit_code == 0
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        "/some/test_code_style_formatting_clean/path/isort",
                        "--profile",
                        "black",
                        ".",
                    ]
                ),
                mocker.call(
                    ["/some/test_code_style_formatting_clean/path/black", "."]
                ),
            ]
        )

    def test_unknown_error(self, click_runner, mocker):
        mocker.patch(
            "build_harness.commands.code_style.run_command",
            side_effect=RuntimeError(),
        )
        result = click_runner.invoke(main, ["formatting"])

        assert result.exit_code == ExitState.UNKNOWN_ERROR.value


class TestCheckArgument:
    def test_isort_fails(self, click_runner, mocker):
        mock_isort_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_isort_result.returncode = 1
        mock_black_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_black_result.returncode = 0
        mocker.patch(
            "build_harness.commands.code_style.run_command",
            side_effect=[mock_isort_result, mock_black_result],
        )
        result = click_runner.invoke(main, ["formatting", "--check"])

        assert result.exit_code == ExitState.ISORT_CHECK_FAILED.value

    def test_black_fails(self, click_runner, mocker):
        mock_isort_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_isort_result.returncode = 0
        mock_black_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_black_result.returncode = 1
        mocker.patch(
            "build_harness.commands.code_style.run_command",
            side_effect=[mock_isort_result, mock_black_result],
        )
        result = click_runner.invoke(main, ["formatting", "--check"])

        assert result.exit_code == ExitState.BLACK_CHECK_FAILED.value

    def test_check_passes(self, click_runner, mocker):
        mock_isort_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_isort_result.returncode = 0
        mock_black_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_black_result.returncode = 0
        mocker.patch(
            "build_harness.commands.code_style.run_command",
            side_effect=[mock_isort_result, mock_black_result],
        )
        result = click_runner.invoke(main, ["formatting", "--check"])

        assert result.exit_code == 0

    def test_check_unknown_error(self, click_runner, mocker):
        mocker.patch(
            "build_harness.commands.code_style.run_command",
            side_effect=RuntimeError(),
        )
        result = click_runner.invoke(main, ["formatting", "--check"])

        assert result.exit_code == ExitState.UNKNOWN_ERROR.value



#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.

import pathlib

import pytest
from {{cookiecutter.project_slug}}._version import DEFAULT_RELEASE_ID, acquire_version


def test_clean(mocker):
    mock_here = mocker.create_autospec(pathlib.Path)

    mocker.patch("{{cookiecutter.project_slug}}._version.pathlib.Path.absolute", return_value=mock_here)
    mock_here.is_file.return_value = True

    mock_handle = mock_here.open.return_value.__enter__.return_value
    mock_handle.read.return_value = "  3.14.159\n"

    result = acquire_version()

    assert result == "3.14.159"


def test_missing_file(mocker):
    mock_here = mocker.create_autospec(pathlib.Path)

    mocker.patch("{{cookiecutter.project_slug}}._version.pathlib.Path.absolute", return_value=mock_here)
    mock_here.is_file.return_value = False

    result = acquire_version()

    assert result == DEFAULT_RELEASE_ID


def test_empty_file(mocker):
    mock_here = mocker.create_autospec(pathlib.Path)

    mocker.patch("{{cookiecutter.project_slug}}._version.pathlib.Path.absolute", return_value=mock_here)
    mock_here.is_file.return_value = True

    mock_handle = mock_here.open.return_value.__enter__.return_value
    mock_handle.read.return_value = ""

    with pytest.raises(RuntimeError, match=r"^Unable to acquire version"):
        acquire_version()



#
#  Copyright (c) 2023 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#
import os.path

import invoke

BUILD_DOCKERFILE = "Dockerfile-apt-lock"


@invoke.task
def generate_lockfile(
    context: invoke.Context,
    target: str,
) -> None:
    lock_file = f"{target}.lock"
    src_file = f"{target}.txt"
    docker_image = f"{target}".replace("_", "-")

    assert os.path.isfile(src_file), f"File {src_file} does not exist"

    # make sure lock file does not exist
    try:
        os.remove(lock_file)
    except OSError:
        pass

    context.run(
        (
            "docker "
            "build "
            f"--build-arg requirements_src={target} "
            "--no-cache "
            f"-t {docker_image} "
            f"-f {BUILD_DOCKERFILE} "
            "."
        ),
        echo=True,
    )
    context.run(
        (
            "docker "
            "run "
            "--rm "
            "-v $(PWD):/media "
            "-w /media "
            f"{docker_image}"
        ),
        echo=True,
    )
    assert os.path.isfile(lock_file), f"File {lock_file} has not been generated"



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import pytest
from parver import ParseError

from build_harness.commands._release_id import (
    InvalidReleaseId,
    validate_release_id,
)


def test_public_release():
    expected_release = "3.1.4"
    result = validate_release_id(expected_release)

    assert result.public == expected_release
    assert result.post is None


def test_replace_post():
    expected_release = "3.1.4"
    result = validate_release_id(expected_release)

    assert result.public == expected_release
    assert result.post is None

    v = result.replace(post=5)
    assert v.post == 5


def test_post_release():
    expected_release = "3.1.4.post11"
    result = validate_release_id(expected_release)

    assert result.public == expected_release
    assert result.post == 11


def test_local_dryrun():
    expected_release = "3.1.4.post11+dryrun"
    result = validate_release_id(expected_release)

    assert result.public == "3.1.4.post11"
    assert result.post == 11
    assert result.local == "dryrun"
    assert str(result) == expected_release


def test_dev_release():
    expected_release = "3.1.4.post11.dev456"
    result = validate_release_id(expected_release)

    assert result.public == "3.1.4.post11.dev456"
    assert result.post == 11
    assert result.dev == 456
    assert str(result) == expected_release


def test_failed_parse_raises(mocker):
    expected_release = "bad_tag"

    mocker.patch(
        "build_harness.commands._release_id.parver.Version.parse",
        side_effect=ParseError("some error"),
    )

    with pytest.raises(
        InvalidReleaseId, match=r"^Tags must be PEP-440 compliant"
    ):
        validate_release_id(expected_release)



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""Actions relating to release ids."""

import logging

import parver  # type: ignore

log = logging.getLogger(__name__)


class InvalidReleaseId(Exception):
    """Parsed release ID is not valid PEP-440."""


def validate_release_id(release_id: str) -> parver.Version:
    """
    Validate release id according to PEP-440.

    https://www.python.org/dev/peps/pep-0440/

    Args:
        release_id: Release id to validate.

    Raises:
        InvalidReleaseId: If release id is not valid PEP-440.
    """
    try:
        return parver.Version.parse(release_id)
    except parver.ParseError as e:
        log.debug("handling parver.ParseError, {0}".format(str(e)))
        message = "Tags must be PEP-440 compliant, {0}".format(release_id)
        raise InvalidReleaseId(message) from e



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

from build_harness.entrypoint import build_harness_entry, release_flow_entry


def test_build_harness_entry(mocker):
    mock_main = mocker.patch("build_harness.entrypoint.main")

    build_harness_entry()

    mock_main.assert_called_once()


def test_release_flow_entry(mocker):
    mock_main = mocker.patch("build_harness.entrypoint.release_flow_main")

    release_flow_entry()

    mock_main.assert_called_once()



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import logging
import pathlib
import subprocess
import tempfile

import pytest

from build_harness.commands import main
from build_harness.commands.publishing import PublishOptions, _publish_packages
from tests.ci.support.click_runner import click_runner  # noqa: F401


@pytest.fixture()
def mock_run(mock_sysargv, mocker):
    mock_pytest_result = mocker.create_autospec(subprocess.CompletedProcess)
    mock_pytest_result.returncode = 0
    mock_pytest_result.stdout = ""
    this_run = mocker.patch(
        "build_harness.commands.publishing.run_command",
        return_value=mock_pytest_result,
    )

    return this_run


class TestPublishPackages:
    MOCK_VENV = pathlib.Path("some/path")

    def test_yes(self, caplog, mock_run, mocker):
        with caplog.at_level(logging.DEBUG):
            _publish_packages(self.MOCK_VENV, PublishOptions.yes, None, None)

            mock_run.assert_has_calls(
                [
                    mocker.call(
                        mocker.ANY,
                        env=dict(),
                        suppress_command_logging=True,
                        suppress_argument_logging=True,
                    )
                ]
            )
            assert "twine command run," in caplog.text

    def test_no(self, caplog, mock_run):
        with caplog.at_level(logging.DEBUG):
            _publish_packages(self.MOCK_VENV, PublishOptions.no, None, None)

            mock_run.assert_not_called()
            assert "Publish disabled" in caplog.text

    def test_dryrun(self, caplog, mock_run):
        with caplog.at_level(logging.DEBUG):
            _publish_packages(self.MOCK_VENV, PublishOptions.dryrun, None, None)

            mock_run.assert_not_called()
            assert "Dry run," in caplog.text

    def test_test(self, caplog, mock_run, mocker):
        with caplog.at_level(logging.DEBUG):
            _publish_packages(self.MOCK_VENV, PublishOptions.test, None, None)

            mock_run.assert_has_calls(
                [
                    mocker.call(
                        mocker.ANY,
                        env=dict(),
                        suppress_command_logging=True,
                        suppress_argument_logging=True,
                    )
                ]
            )
            assert "twine command run," in caplog.text

    def test_none_user_password(self, caplog, mock_run, mocker):
        with caplog.at_level(logging.DEBUG):
            _publish_packages(self.MOCK_VENV, PublishOptions.yes, None, None)

            mock_run.assert_has_calls(
                [
                    mocker.call(
                        mocker.ANY,
                        env=dict(),
                        suppress_command_logging=True,
                        suppress_argument_logging=True,
                    )
                ]
            )

            assert "twine username not provided" in caplog.text
            assert "twine password not provided" in caplog.text

    def test_user(self, caplog, mock_run, mocker):
        with caplog.at_level(logging.DEBUG):
            _publish_packages(
                self.MOCK_VENV, PublishOptions.yes, None, "this_user"
            )

            mock_run.assert_has_calls(
                [
                    mocker.call(
                        mocker.ANY,
                        env={
                            "TWINE_USERNAME": "this_user",
                        },
                        suppress_command_logging=True,
                        suppress_argument_logging=True,
                    )
                ]
            )
            assert "twine username," in caplog.text
            assert "twine password," not in caplog.text

    def test_password(self, caplog, mock_run, mocker):
        with caplog.at_level(logging.DEBUG):
            _publish_packages(
                self.MOCK_VENV, PublishOptions.yes, "this_password", None
            )

            mock_run.assert_has_calls(
                [
                    mocker.call(
                        mocker.ANY,
                        env={
                            "TWINE_PASSWORD": "this_password",
                        },
                        suppress_command_logging=True,
                        suppress_argument_logging=True,
                    )
                ]
            )
            assert "twine username," not in caplog.text
            assert "twine password," in caplog.text

    def test_user_password(self, caplog, mock_run, mocker):
        with caplog.at_level(logging.DEBUG):
            _publish_packages(
                self.MOCK_VENV, PublishOptions.yes, "this_password", "this_user"
            )

            mock_run.assert_has_calls(
                [
                    mocker.call(
                        mocker.ANY,
                        env={
                            "TWINE_USERNAME": "this_user",
                            "TWINE_PASSWORD": "this_password",
                        },
                        suppress_command_logging=True,
                        suppress_argument_logging=True,
                    )
                ]
            )
            assert "twine username," in caplog.text
            assert "twine password," in caplog.text


class TestPublish:
    def test_default(self, caplog, click_runner, mocker, mock_run):
        with caplog.at_level(logging.DEBUG):
            result = click_runner.invoke(
                main,
                [
                    "--log-console-enable",
                    "--log-level",
                    "debug",
                    "publish",
                    "--user",
                    "username",
                    "--password",
                    "password",
                ],
            )

            assert result.exit_code == 0
            mock_run.assert_has_calls(
                [
                    mocker.call(
                        [
                            "/some/conftest/path/twine",
                            "upload",
                            "--non-interactive",
                            "dist/*.whl",
                            "dist/*.tar.gz",
                        ],
                        env={
                            "TWINE_USERNAME": "username",
                            "TWINE_PASSWORD": "password",
                        },
                        suppress_command_logging=True,
                        suppress_argument_logging=True,
                    ),
                ]
            )

            assert "twine command run," in caplog.text
            assert "twine username," in caplog.text
            assert "twine password," in caplog.text
            assert "password passed as CLI argument," in caplog.text
            # check that run_command logging is suppressed to protect credentials
            assert "command to run," not in caplog.text
            assert "command arguments," not in caplog.text

    def test_publish(self, click_runner, mocker, mock_run):
        result = click_runner.invoke(
            main,
            [
                "publish",
                "--user",
                "username",
                "--password",
                "password",
                "--publish",
                "yes",
            ],
        )

        assert result.exit_code == 0
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        "/some/conftest/path/twine",
                        "upload",
                        "--non-interactive",
                        "dist/*.whl",
                        "dist/*.tar.gz",
                    ],
                    env={
                        "TWINE_PASSWORD": "password",
                        "TWINE_USERNAME": "username",
                    },
                    suppress_command_logging=True,
                    suppress_argument_logging=True,
                ),
            ]
        )

    def test_nopublish(self, click_runner, mock_run):
        result = click_runner.invoke(
            main,
            [
                "publish",
                "--user",
                "username",
                "--password",
                "password",
                "--publish",
                "no",
            ],
        )

        assert result.exit_code == 0
        mock_run.assert_not_called()

    def test_publish_dryrun(self, click_runner, mock_run):
        """dryrun overrides publish"""
        result = click_runner.invoke(
            main,
            [
                "publish",
                "--user",
                "username",
                "--password",
                "password",
                "--publish",
                "dryrun",
            ],
        )

        assert result.exit_code == 0
        mock_run.assert_not_called()

    def test_password_file(self, caplog, click_runner, mock_run, mocker):
        with caplog.at_level(
            logging.DEBUG
        ), tempfile.TemporaryDirectory() as dir:
            this_dir = pathlib.Path(dir)
            this_file = this_dir / "this-file"
            with this_file.open(mode="w") as f:
                f.write("this-password")

            result = click_runner.invoke(
                main,
                [
                    "--log-console-enable",
                    "--log-level",
                    "debug",
                    "publish",
                    "--user",
                    "username",
                    "--password-file",
                    str(this_file),
                ],
            )

            assert result.exit_code == 0

            mock_run.assert_has_calls(
                [
                    mocker.call(
                        mocker.ANY,
                        env={
                            "TWINE_USERNAME": "username",
                            "TWINE_PASSWORD": "this-password",
                        },
                        suppress_command_logging=True,
                        suppress_argument_logging=True,
                    )
                ]
            )
            assert "password specified in password file," in caplog.text
            assert "password recovered from file," in caplog.text



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

from build_harness import __version__
from build_harness.commands import main
from tests.ci.support.click_runner import click_runner  # noqa: F401


def test_main(click_runner):
    result = click_runner.invoke(main, [])

    assert result.exit_code == 0


def test_help(click_runner):
    result = click_runner.invoke(main, ["--help"])

    assert result.exit_code == 0
    assert "Usage: main [OPTIONS] COMMAND [ARGS]" in result.output


def test_version(click_runner):
    result = click_runner.invoke(main, ["--version"])

    assert result.exit_code == 0
    assert __version__ in result.output



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import copy
import pathlib
import subprocess
import typing

import pytest

from build_harness.commands import main
from build_harness.commands.dependencies import (
    DependencyCheckError,
    DependencyInstallError,
    ExitState,
    ProjectDependencies,
    PyprojecttomlError,
    VenvError,
    VenvPipError,
    _check_dependencies,
    _check_dependency_names,
    _check_virtual_environment,
    _extract_pip_list_packages,
    _install_all_project_dependencies,
    _install_project_dependencies,
)
from tests.ci.support.click_runner import click_runner  # noqa: F401

MOCK_VENV_BIN_PATH = pathlib.Path("some/test_dependencies/bin")


class TestExtractPipListPackages:
    sample_output = """
Package                       Version
----------------------------- ---------
alabaster                     0.7.12
black                         20.8b1
certifi                       2020.6.20
"""

    def test_clean(self):
        result = _extract_pip_list_packages(self.sample_output)

        assert result == [
            ("alabaster", "0.7.12"),
            ("black", "20.8b1"),
            ("certifi", "2020.6.20"),
        ]

    def test_empty(self):
        sample_output = """"""

        result = _extract_pip_list_packages(sample_output)

        assert result == list()

    def test_empty_list(self):
        sample_output = """
Package                       Version
----------------------------- ---------
"""

        result = _extract_pip_list_packages(sample_output)

        assert result == list()


class TestCheckDependencies:
    def test_clean(self, mocker):
        mock_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_result.returncode = 0
        mock_result.stdout = """
Package                       Version
----------------------------- ---------
pkg1                     0.7.12
pkg_name2                         20.8b1
pkg3                       2020.6.20
"""
        mock_dependencies = ProjectDependencies(
            runtime=["pkg-name2==20.8b1"], dev=list(), doc=list(), test=list()
        )

        mocker.patch(
            "build_harness.commands.dependencies.run_command",
            return_value=mock_result,
        )

        _check_dependencies(MOCK_VENV_BIN_PATH, mock_dependencies, "all")

    def test_empty_venv_dependencies(self, mocker):
        """Empty venv with project dependencies fails."""
        mock_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_result.returncode = 0
        mock_result.stdout = """
Package                       Version
----------------------------- ---------
"""
        mock_dependencies = ProjectDependencies(
            runtime=["pkg2==20.8b1"], dev=list(), doc=list(), test=list()
        )

        mocker.patch(
            "build_harness.commands.dependencies.run_command",
            return_value=mock_result,
        )

        with pytest.raises(
            DependencyCheckError,
            match=r"^Project dependencies not installed",
        ):
            _check_dependencies(MOCK_VENV_BIN_PATH, mock_dependencies, "all")

    def test_bad_specifier(self, mocker):
        mock_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_result.returncode = 0
        mock_result.stdout = """
Package                       Version
----------------------------- ---------
pkg1                     0.7.12
pkg2                         20.8b1
pkg3                       2020.6.20
"""
        mock_dependencies = ProjectDependencies(
            runtime=["pkg2==19.0"], dev=list(), doc=list(), test=list()
        )

        mocker.patch(
            "build_harness.commands.dependencies.run_command",
            return_value=mock_result,
        )

        with pytest.raises(
            DependencyCheckError,
            match=r"^Installed packages do not comply with declared project dependencies",
        ):
            _check_dependencies(MOCK_VENV_BIN_PATH, mock_dependencies, "all")

    def test_missing_package(self, mocker):
        mock_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_result.returncode = 0
        mock_result.stdout = """
Package                       Version
----------------------------- ---------
pkg1                     0.7.12
pkg2                         20.8b1
pkg3                       2020.6.20
"""
        mock_dependencies = ProjectDependencies(
            runtime=["pkg4==3.1.4"], dev=list(), doc=list(), test=list()
        )

        mocker.patch(
            "build_harness.commands.dependencies.run_command",
            return_value=mock_result,
        )

        with pytest.raises(
            DependencyCheckError,
            match=r"^Project dependencies not installed",
        ):
            _check_dependencies(MOCK_VENV_BIN_PATH, mock_dependencies, "all")

    def test_empty_dependencies(self, mocker):
        """No declared dependencies passes."""
        mock_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_result.returncode = 0
        mock_result.stdout = """
Package                       Version
----------------------------- ---------
"""
        mock_dependencies = ProjectDependencies(
            runtime=[], dev=list(), doc=list(), test=list()
        )

        mocker.patch(
            "build_harness.commands.dependencies.run_command",
            return_value=mock_result,
        )

        _check_dependencies(MOCK_VENV_BIN_PATH, mock_dependencies, "all")

    def test_bad_venv_raises(self, mocker):
        mock_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_result.returncode = 1
        mock_dependencies = ProjectDependencies(
            runtime=["a==3.1.4"], dev=list(), doc=list(), test=list()
        )

        mocker.patch(
            "build_harness.commands.dependencies.run_command",
            return_value=mock_result,
        )

        with pytest.raises(
            VenvPipError,
            match=r"^Failed to list installed virtual environment packages",
        ):
            _check_dependencies(MOCK_VENV_BIN_PATH, mock_dependencies, "all")


@pytest.fixture()
def mock_venv_create(mocker):
    def _apply(exists: bool) -> None:
        mocker.patch(
            "build_harness.commands.dependencies.pathlib.Path.exists",
            return_value=exists,
        )
        if not exists:
            # mock out the command that creates a venv
            mock_result = mocker.create_autospec(subprocess.CompletedProcess)
            mock_result.returncode = 0
            mock_result.stdout = """
some status
"""
            mock_result.stderr = ""
            mocker.patch(
                "build_harness.commands.dependencies.run_command",
                return_value=mock_result,
            )

    return _apply


class TestCheckVirtualEnvironment:
    def test_clean_create(self, mock_venv_create):
        mock_venv_create(False)

        result = _check_virtual_environment(MOCK_VENV_BIN_PATH)

        assert result == MOCK_VENV_BIN_PATH / ".venv" / "bin"

    def test_clean_exists(self, mocker, mock_venv_create):
        mock_venv_create(True)

        mocker.patch(
            "build_harness.commands.dependencies.pathlib.Path.is_file",
            return_value=True,
        )

        result = _check_virtual_environment(MOCK_VENV_BIN_PATH)

        assert result == MOCK_VENV_BIN_PATH / ".venv" / "bin"

    def test_command_fails(self, mocker, mock_venv_create):
        mock_venv_create(False)
        mocker.patch(
            "build_harness.commands.dependencies.pathlib.Path.exists",
            return_value=False,
        )
        mock_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_result.returncode = 1
        mock_result.stdout = """
some status
"""
        mock_result.stderr = """
command failed
"""
        mocker.patch(
            "build_harness.commands.dependencies.run_command",
            return_value=mock_result,
        )

        with pytest.raises(
            VenvError, match=r"^Failed creating virtual environment"
        ):
            _check_virtual_environment(MOCK_VENV_BIN_PATH)

    def test_bad_venv_bin_raises(self, mocker, mock_venv_create):
        mock_venv_create(True)

        mocker.patch(
            "build_harness.commands.dependencies.pathlib.Path.is_file",
            return_value=False,
        )

        with pytest.raises(VenvError, match=r"^Invalid virtual environment"):
            _check_virtual_environment(MOCK_VENV_BIN_PATH)


class TestInstallProjectDependencies:
    def test_default(self, mocker):
        mock_dependencies = [
            "toml == 3.1.4",
            "flit >=2.0,<4",
        ]
        mock_command = mocker.patch(
            "build_harness.commands.dependencies.run_command"
        )
        mock_command.return_value.returncode = 0

        _install_project_dependencies(mock_dependencies, MOCK_VENV_BIN_PATH)

        expected_stdin = "\n".join(mock_dependencies)
        mock_command.assert_called_once_with(
            [
                str(MOCK_VENV_BIN_PATH / "pip"),
                "install",
                "-r",
                "/dev/stdin",
            ],
            input=expected_stdin,
            text=True,
            capture_output=True,
            universal_newlines=True,
        )

    def test_install_raises(self, mocker):
        mock_dependencies = [
            "toml == 3.1.4",
            "flit >=2.0,<4",
        ]
        mock_command = mocker.patch(
            "build_harness.commands.dependencies.run_command"
        )
        mock_command.return_value.returncode = 1

        with pytest.raises(
            DependencyInstallError, match=r"^Dependency installation failed"
        ):
            _install_project_dependencies(mock_dependencies, MOCK_VENV_BIN_PATH)


class TestInstallAllProjectDependencies:
    def test_all(self, mocker):
        mock_project_dependencies = {
            "runtime": [
                "requests >=2.2, <3.0",
                "toml == 1.0.0",
            ],
            "dev": ["pre_commit == 2.7.1"],
            "doc": ["sphinx == 3.2.1"],
            "test": ["pytest == 6.1.1"],
        }
        mock_venv = pathlib.Path("/some/venv")
        mock_install = mocker.patch(
            "build_harness.commands.dependencies._install_project_dependencies"
        )

        _install_all_project_dependencies(mock_project_dependencies, mock_venv)

        mock_install.assert_called_once_with(
            [
                "requests >=2.2, <3.0",
                "toml == 1.0.0",
                "pre_commit == 2.7.1",
                "sphinx == 3.2.1",
                "pytest == 6.1.1",
            ],
            mock_venv,
        )


@pytest.fixture()
def mock_run(mock_sysargv, mocker):
    mock_pytest_result = mocker.create_autospec(subprocess.CompletedProcess)
    mock_pytest_result.returncode = 0
    mock_pytest_result.stdout = ""
    this_run = mocker.patch(
        "build_harness.commands.dependencies.run_command",
        return_value=mock_pytest_result,
    )

    return this_run


MOCK_DEPENDENCIES: ProjectDependencies = {
    "runtime": [
        "requests == 3.1.4",
    ],
    "dev": [
        "behave == 1.2.6",
    ],
    "doc": [
        "sphinx == 3.2.1",
    ],
    "test": [
        "pytest == 6.1.1",
    ],
}


@pytest.fixture()
def mock_dependencies(mocker):
    def _apply(venv_bin_path: pathlib.Path) -> typing.Any:
        mocker.patch(
            "build_harness.commands.dependencies.acquire_project_dependencies",
            return_value=copy.deepcopy(MOCK_DEPENDENCIES),
        )
        mock_check = mocker.patch(
            "build_harness.commands.dependencies._check_virtual_environment",
            return_value=venv_bin_path,
        )

        return mock_check

    return _apply


class TestInstall:
    def test_default(self, click_runner, mocker, mock_dependencies, mock_run):
        mock_check = mock_dependencies(MOCK_VENV_BIN_PATH)

        result = click_runner.invoke(
            main,
            [
                "install",
            ],
        )

        assert result.exit_code == 0
        mock_check.assert_called_once()
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        str(MOCK_VENV_BIN_PATH / "pip"),
                        "install",
                        "-r",
                        "/dev/stdin",
                    ],
                    capture_output=True,
                    input="requests == 3.1.4\nbehave == 1.2.6\nsphinx == "
                    "3.2.1\npytest == 6.1.1",
                    text=True,
                    universal_newlines=True,
                ),
            ]
        )

    def test_all(self, click_runner, mocker, mock_dependencies, mock_run):
        mock_check = mock_dependencies(MOCK_VENV_BIN_PATH)

        result = click_runner.invoke(
            main,
            [
                "install",
                "--dependencies",
                "all",
            ],
        )

        assert result.exit_code == 0
        mock_check.assert_called_once()
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        str(MOCK_VENV_BIN_PATH / "pip"),
                        "install",
                        "-r",
                        "/dev/stdin",
                    ],
                    capture_output=True,
                    input="requests == 3.1.4\nbehave == 1.2.6\nsphinx == 3.2.1\npytest == 6.1.1",
                    text=True,
                    universal_newlines=True,
                ),
            ]
        )

    def test_runtime(self, click_runner, mocker, mock_dependencies, mock_run):
        mock_check = mock_dependencies(MOCK_VENV_BIN_PATH)

        result = click_runner.invoke(
            main,
            [
                "install",
                "--dependencies",
                "runtime",
            ],
        )

        assert result.exit_code == 0
        mock_check.assert_called_once()
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        str(MOCK_VENV_BIN_PATH / "pip"),
                        "install",
                        "-r",
                        "/dev/stdin",
                    ],
                    capture_output=True,
                    input="requests == 3.1.4",
                    text=True,
                    universal_newlines=True,
                ),
            ]
        )

    def test_dev(self, click_runner, mocker, mock_dependencies, mock_run):
        mock_check = mock_dependencies(MOCK_VENV_BIN_PATH)

        result = click_runner.invoke(
            main,
            [
                "install",
                "--dependencies",
                "dev",
            ],
        )

        assert result.exit_code == 0
        mock_check.assert_called_once()
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        str(MOCK_VENV_BIN_PATH / "pip"),
                        "install",
                        "-r",
                        "/dev/stdin",
                    ],
                    capture_output=True,
                    input="behave == 1.2.6",
                    text=True,
                    universal_newlines=True,
                ),
            ]
        )

    def test_doc(self, click_runner, mocker, mock_dependencies, mock_run):
        mock_check = mock_dependencies(MOCK_VENV_BIN_PATH)

        result = click_runner.invoke(
            main,
            [
                "install",
                "--dependencies",
                "doc",
            ],
        )

        assert result.exit_code == 0
        mock_check.assert_called_once()
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        str(MOCK_VENV_BIN_PATH / "pip"),
                        "install",
                        "-r",
                        "/dev/stdin",
                    ],
                    capture_output=True,
                    input="sphinx == 3.2.1",
                    text=True,
                    universal_newlines=True,
                ),
            ]
        )

    def test_test(self, click_runner, mocker, mock_dependencies, mock_run):
        mock_check = mock_dependencies(MOCK_VENV_BIN_PATH)

        result = click_runner.invoke(
            main,
            [
                "install",
                "--dependencies",
                "test",
            ],
        )

        assert result.exit_code == 0
        mock_check.assert_called_once()
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        str(MOCK_VENV_BIN_PATH / "pip"),
                        "install",
                        "-r",
                        "/dev/stdin",
                    ],
                    capture_output=True,
                    input="pytest == 6.1.1",
                    text=True,
                    universal_newlines=True,
                ),
            ]
        )

    def test_bad_pyprojecttoml(self, click_runner, mocker, mock_dependencies):
        mock_dependencies(MOCK_VENV_BIN_PATH)

        mocker.patch(
            "build_harness.commands.dependencies.acquire_project_dependencies",
            side_effect=PyprojecttomlError("Missing pyproject.toml file"),
        )
        result = click_runner.invoke(main, ["install"])

        assert result.exit_code == ExitState.BAD_PYPROJECTTOML.value
        assert "Missing pyproject.toml file" in result.output

    def test_fails(self, click_runner, mocker, mock_dependencies):
        mock_dependencies(MOCK_VENV_BIN_PATH)

        mocker.patch(
            "build_harness.commands.dependencies._install_project_dependencies",
            side_effect=DependencyInstallError(
                "Dependency installation failed"
            ),
        )
        result = click_runner.invoke(main, ["install"])

        assert result.exit_code == ExitState.DEPENDENCY_INSTALL_FAILED.value
        assert "Dependency installation failed" in result.output


class TestCheckArgument:
    def test_clean(self, click_runner, mocker, mock_dependencies):
        mock_dependencies(MOCK_VENV_BIN_PATH)

        mock_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_result.returncode = 0
        mock_result.stdout = """
Package                       Version
----------------------------- ---------
pkg1                     0.7.12
pkg2                         20.8b1
requests   3.1.4
behave   1.2.6
sphinx   3.2.1
pytest   6.1.1
"""
        mocker.patch(
            "build_harness.commands.dependencies.run_command",
            return_value=mock_result,
        )
        mocker.patch(
            "build_harness.commands.dependencies.acquire_project_dependencies",
            return_value=MOCK_DEPENDENCIES,
        )
        result = click_runner.invoke(main, ["install", "--check"])

        assert result.exit_code == 0

    def test_check_fails(self, click_runner, mocker, mock_dependencies):
        mock_dependencies(MOCK_VENV_BIN_PATH)

        mocker.patch(
            "build_harness.commands.dependencies._check_dependencies",
            side_effect=DependencyCheckError("Dependency check failed"),
        )
        result = click_runner.invoke(main, ["install", "--check"])

        assert result.exit_code == ExitState.DEPENDENCY_CHECK_FAILED.value
        assert "Dependency check failed" in result.output

    def test_pip_fails(self, click_runner, mocker, mock_dependencies):
        mock_dependencies(MOCK_VENV_BIN_PATH)

        mock_result = mocker.create_autospec(subprocess.CompletedProcess)
        mock_result.returncode = 1
        mocker.patch(
            "build_harness.commands.dependencies.run_command",
            return_value=mock_result,
        )
        result = click_runner.invoke(main, ["install", "--check"])

        assert result.exit_code == ExitState.DEPENDENCY_CHECK_FAILED.value
        assert (
            "Failed to list installed virtual environment packages"
            in result.output
        )



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import contextlib
import os
import pathlib
import tempfile


@contextlib.contextmanager
def change_directory():
    current_dir = pathlib.Path(os.curdir).absolute()
    with tempfile.TemporaryDirectory() as this_dir:
        dir_path = pathlib.Path(this_dir)
        os.chdir(dir_path)

        yield dir_path

    os.chdir(current_dir)



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""Define exit state numbers for build harness CLI."""

import dataclasses
import enum
import pathlib


@dataclasses.dataclass
class CommandState:
    """Global click command state context."""

    project_path: pathlib.Path
    venv_path: pathlib.Path


@enum.unique
class ExitState(enum.Enum):
    """Exit state enumeration."""

    UNKNOWN_ERROR = 100

    FORMATTING_FAILED = 110
    ISORT_CHECK_FAILED = 111
    BLACK_CHECK_FAILED = 112

    STATIC_ANALYSIS_FAILED = 120
    FLAKE8_FAILED = 121
    MYPY_FAILED = 122

    TESTS_FAILED = 130
    TEST_COVERAGE_FAILED = 131

    ACCEPTANCE_RUN_FAILED = 140
    ACCEPTANCE_TESTS_FAILED = 141

    DEPENDENCY_INSTALL_FAILED = 150
    DEPENDENCY_CHECK_FAILED = 151

    PACKAGING_FAILED = 160

    PUBLISHING_FAILED = 170

    BAD_VERSION = 180
    BAD_REPO = 190
    BAD_GIT_EXE = 200
    BAD_PYPROJECTTOML = 210

    BOOTSTRAP_FAILED = 220



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""Release flow command entry point."""

import logging
import pathlib
import sys
import traceback
import typing

import click
from click_logging_config import LoggingConfiguration, logging_parameters

from build_harness._utility import report_console_error
from build_harness._version import __version__, acquire_version
from build_harness.tools.git import (
    DEFAULT_DEFAULT_BRANCH_NAME,
    GitNotFoundError,
    GitRepoError,
    get_tag_data,
    validate_git,
)

from ._declarations import (
    DEFAULT_CONSOLE_LOGGING_ENABLED,
    DEFAULT_FILE_LOGGING_ENABLED,
    DEFAULT_PROJECT_PATH,
)
from ._release_id import InvalidReleaseId, validate_release_id
from .state import CommandState, ExitState

log = logging.getLogger(__name__)

DEFAULT_RELEASEFLOW_LOGCONFIG = LoggingConfiguration.parse_obj(
    {
        "enable_file_logging": DEFAULT_FILE_LOGGING_ENABLED,
        "enable_console_logging": DEFAULT_CONSOLE_LOGGING_ENABLED,
        "file_logging": {"log_file_path": pathlib.Path("release_flow.log")},
        "log_level": "debug",
    }
)


@click.command()
@click.version_option(version=acquire_version())
@click.option(
    "--default-branch",
    default=DEFAULT_DEFAULT_BRANCH_NAME,
    help="Git default/main branch name.",
    show_default=True,
    type=str,
)
@click.option(
    "--git",
    default=None,
    help="Git executable. [default: users PATH]",
    type=str,
)
@click.option(
    "--project",
    default=DEFAULT_PROJECT_PATH,
    help="Git project directory.",
    show_default=True,
    type=str,
)
@logging_parameters(DEFAULT_RELEASEFLOW_LOGCONFIG)
def release_flow_main(
    default_branch: str, git: typing.Optional[str], project: str
) -> None:
    """
    Construct a release id from git tag and commit history.

    If HEAD commit is tagged then use the tag as the release id. Otherwise try
    to count the number of commits since the most recent tag, with a PEP-440
    constructed release id of the form ``<tag>.post<N>``. Note that this
    implies that the tag must be PEP-440 compliant.

    If there are no tags in commit history then a default tag of ``0.0.0`` is
    inferred and the number of commits counted since the root (first) commit in
    commit history.

    The count method is slightly different when there are no tags in commit
    history so it is recommended to tag the root commit with ``0.0.0`` so that
    the count method is always the same.
    """
    try:
        state = CommandState(
            project_path=pathlib.Path(project),
            venv_path=pathlib.Path(sys.argv[0]).parent.absolute(),
        )
        log.info(f"build harness version, {__version__}")

        validate_git(git)

        tag_data = get_tag_data(state.project_path, default_branch)
        tag_version = validate_release_id(tag_data.tag)
        if tag_data.offset:
            updated_version = tag_version.replace(
                post=int(tag_data.offset), post_sep1="-", post_sep2="."
            )
            release_id = str(updated_version)
        else:
            release_id = str(tag_version)

        click.echo(release_id, err=False, nl=False)
    except InvalidReleaseId as e:
        report_console_error(str(e))
        sys.exit(ExitState.BAD_VERSION.value)
    except GitRepoError as e:
        report_console_error(str(e))
        sys.exit(ExitState.BAD_REPO.value)
    except GitNotFoundError as e:
        report_console_error(str(e))
        click.echo("Git must be installed and configured to use this utility.")
        click.echo("In Linux, something like this is necessary:")
        click.echo("    sudo apt install -y git || sudo yum install git")
        click.echo('    git config user.name "Your Name"')
        click.echo('    git config user.email "you@example.com"')
        sys.exit(ExitState.BAD_GIT_EXE.value)
    except Exception as e:
        message = "Unexpected error. Check log for details."
        log.exception(str(e))
        report_console_error(message)
        report_console_error(str(e))
        stack_trace = traceback.format_exc()
        report_console_error(stack_trace)
        sys.exit(ExitState.UNKNOWN_ERROR.value)



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import copy
import pathlib

from build_harness.tools import CommandOptionsTestTool, PytestCommand

MOCK_OPTIONS: CommandOptionsTestTool = {
    "output_path": pathlib.Path("dist"),
    "source_path": pathlib.Path("."),
    "test_path": pathlib.Path("tests"),
    "venv_path": pathlib.Path("some/venv/path"),
    "report_enabled": {
        "junitxml": False,
        "term-missing": False,
        "html": False,
        "xml": False,
    },
    "report_dirs": {
        "junitxml": None,
        "term-missing": None,
        "html": None,
        "xml": None,
    },
}


def test_default():
    this_options = copy.deepcopy(MOCK_OPTIONS)
    under_test = PytestCommand(this_options)

    assert under_test.command() == [
        "some/venv/path/pytest",
        "tests",
    ]


def test_source_dir():
    this_options = copy.deepcopy(MOCK_OPTIONS)
    this_options["source_path"] = pathlib.Path("some/source/path")
    under_test = PytestCommand(this_options)

    assert under_test.command() == [
        "some/venv/path/pytest",
        "tests",
    ]


def test_test_dir():
    this_options = copy.deepcopy(MOCK_OPTIONS)
    this_options["test_path"] = pathlib.Path("some/test/path")
    under_test = PytestCommand(this_options)

    assert under_test.command() == [
        "some/venv/path/pytest",
        "some/test/path",
    ]


def test_cov_term():
    this_options = copy.deepcopy(MOCK_OPTIONS)
    this_options["report_enabled"]["term-missing"] = True
    under_test = PytestCommand(this_options)

    assert under_test.command() == [
        "some/venv/path/pytest",
        "--cov=.",
        "--cov-report",
        "term-missing",
        "tests",
    ]


def test_cov_html_file():
    this_options = copy.deepcopy(MOCK_OPTIONS)
    this_options["report_enabled"]["html"] = True
    this_options["report_dirs"]["html"] = pathlib.Path("some/html/path")
    under_test = PytestCommand(this_options)

    assert under_test.command() == [
        "some/venv/path/pytest",
        "--cov=.",
        "--cov-report",
        "html:some/html/path",
        "tests",
    ]


def test_cov_xml_file():
    this_options = copy.deepcopy(MOCK_OPTIONS)
    this_options["report_enabled"]["xml"] = True
    this_options["report_dirs"]["xml"] = pathlib.Path("some/xml/path")
    under_test = PytestCommand(this_options)

    assert under_test.command() == [
        "some/venv/path/pytest",
        "--cov=.",
        "--cov-report",
        "xml:some/xml/path",
        "tests",
    ]


def test_junit_xml_file():
    this_options = copy.deepcopy(MOCK_OPTIONS)
    this_options["report_enabled"]["junitxml"] = True
    this_options["report_dirs"]["junitxml"] = pathlib.Path("some/junit/path")
    under_test = PytestCommand(this_options)

    assert under_test.command() == [
        "some/venv/path/pytest",
        "--cov=.",
        "--junitxml=some/junit/path",
        "tests",
    ]


def test_multiple_reports():
    this_options = copy.deepcopy(MOCK_OPTIONS)
    this_options["report_enabled"]["xml"] = True
    this_options["report_enabled"]["html"] = True
    under_test = PytestCommand(this_options)

    assert under_test.command() == [
        "some/venv/path/pytest",
        "--cov=.",
        "--cov-report",
        "html",
        "--cov-report",
        "xml",
        "tests",
    ]



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""Formatting subcommand implementation."""

import copy
import logging
import pathlib
import sys

import click

from build_harness._utility import (
    CommandArgs,
    command_path,
    report_console_error,
    run_command,
)

from .state import CommandState, ExitState

BLACK_FORMAT_CMD: CommandArgs = ["black", "."]
ISORT_FORMAT_CMD: CommandArgs = ["isort", "--profile", "black", "."]

BLACK_CHECK_CMD: CommandArgs = ["black", "--check", "."]
ISORT_CHECK_CMD: CommandArgs = ["isort", "--profile", "black", "--diff", "."]

log = logging.getLogger(__name__)


class FormattingError(Exception):
    """Problem occurred during code formatting."""


class IsortCheckError(Exception):
    """The isort formatting check failed."""


class BlackCheckError(Exception):
    """The black formatting check failed."""


def _apply_formatting(venv_path: pathlib.Path) -> None:
    """
    Apply formatting to source code.

    Args:
        venv_path: Path to Python virtual environment.

    Raises:
        FormattingError: If either isort or black exit non-zero.
    """
    click.echo("isort formatting...")
    isort_cmd = copy.deepcopy(ISORT_FORMAT_CMD)
    isort_cmd[0] = command_path(venv_path, isort_cmd)
    isort_result = run_command(isort_cmd)
    click.echo(click.style("isort completed", fg="green"))

    click.echo("black formatting...")
    black_cmd = copy.deepcopy(BLACK_FORMAT_CMD)
    black_cmd[0] = command_path(venv_path, black_cmd)
    black_result = run_command(black_cmd)
    click.echo(click.style("black completed", fg="green"))

    if any(x != 0 for x in [isort_result.returncode, black_result.returncode]):
        raise FormattingError("isort and/or black failed during formatting.")


def _check_formatting(venv_path: pathlib.Path) -> None:
    """
    Check if any formatting changes are needed on the code base.

    Args:
        venv_path: Path to Python virtual environment.

    Raises:
        BlackCheckError: If black formatting check fails.
        IsortCheckError: If isort formatting check fails.
    """
    click.echo("isort formatting check...")
    isort_cmd = copy.deepcopy(ISORT_CHECK_CMD)
    isort_cmd[0] = command_path(venv_path, isort_cmd)
    isort_result = run_command(isort_cmd)
    if isort_result.returncode != 0:
        raise IsortCheckError("isort check failed")
    click.echo(click.style("isort check completed", fg="green"))

    click.echo("black formatting check...")
    black_cmd = copy.deepcopy(BLACK_CHECK_CMD)
    black_cmd[0] = command_path(venv_path, black_cmd)
    black_result = run_command(black_cmd)
    if black_result.returncode != 0:
        raise BlackCheckError("black check failed")
    click.echo(click.style("black check completed", fg="green"))


@click.command()
@click.pass_context
@click.option(
    "--check",
    help="Don't apply formatting, just check if formatting is needed.",
    is_flag=True,
)
def formatting(ctx: click.Context, check: bool) -> None:
    """
    Run suite of Python code formatting.

    The suite currently comprises "isort and "black". isort is run before black
    to resolve cases where isort and black compete over formatting changes.
    """
    try:
        ctx.ensure_object(dict)
        command_state: CommandState = ctx.obj["command_state"]
        if not check:
            _apply_formatting(command_state.venv_path)
        else:
            _check_formatting(command_state.venv_path)
        click.echo(click.style("formatting completed", fg="green"))
    except FormattingError as e:
        report_console_error(str(e))
        sys.exit(ExitState.FORMATTING_FAILED.value)
    except BlackCheckError as e:
        report_console_error(str(e))
        sys.exit(ExitState.BLACK_CHECK_FAILED.value)
    except IsortCheckError as e:
        report_console_error(str(e))
        sys.exit(ExitState.ISORT_CHECK_FAILED.value)
    except Exception:
        message = "Unexpected error. Check log for details."
        log.exception(message)
        report_console_error(message)
        sys.exit(ExitState.UNKNOWN_ERROR.value)



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import contextlib
import logging
import os
import pathlib
import typing

from build_harness.commands import main
from tests.ci.support.click_runner import click_runner  # noqa: F401
from tests.ci.support.project_context import (
    build_cookiecutter_context,
    make_project_venv,
)

log = logging.getLogger(__name__)


@contextlib.contextmanager
def managed_dependencies(
    project_name: str,
) -> typing.Generator[pathlib.Path, None, None]:
    cwd = pathlib.Path.cwd()
    with build_cookiecutter_context() as this_context:
        make_project_venv(project_name, this_context.temp_path)
        os.chdir(this_context.temp_path)
        yield

    os.chdir(cwd)


class TestBootstrap:
    PROJECT_NAME = "my_new_project"
    COMMAND_BASE = ["bootstrap", "--disable-dependencies", "my_new_project"]

    def test_default(self, click_runner):
        with managed_dependencies(self.PROJECT_NAME):
            result = click_runner.invoke(main, self.COMMAND_BASE.copy())

            if result.exit_code != 0:
                print(result.output)
            assert result.exit_code == 0
            assert pathlib.Path("my_new_project").is_dir()

    def test_ci(self, click_runner):
        with managed_dependencies(self.PROJECT_NAME):
            result = click_runner.invoke(
                main, (self.COMMAND_BASE.copy() + ["--ci", "gitlabci"])
            )

            if result.exit_code != 0:
                print(result.output)
            assert result.exit_code == 0
            assert pathlib.Path("my_new_project").is_dir()

    def test_packager(self, click_runner):
        with managed_dependencies(self.PROJECT_NAME):
            result = click_runner.invoke(
                main, (self.COMMAND_BASE.copy() + ["--packager", "flit"])
            )

            if result.exit_code != 0:
                print(result.output)
            assert result.exit_code == 0
            assert pathlib.Path("my_new_project").is_dir()

    def test_default_branch(self, click_runner):
        with click_runner.isolated_filesystem():
            result = click_runner.invoke(
                main, (self.COMMAND_BASE.copy() + ["--default-branch", "main"])
            )

            if result.exit_code != 0:
                print(result.output)
            assert result.exit_code == 0
            assert pathlib.Path("my_new_project").is_dir()

    def test_force(self, click_runner):
        with managed_dependencies(self.PROJECT_NAME):
            result = click_runner.invoke(
                main, (self.COMMAND_BASE.copy() + ["--force"])
            )

            if result.exit_code != 0:
                print(result.output)
            assert result.exit_code == 0
            assert pathlib.Path("my_new_project").is_dir()



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""Python release specific imports."""

import sys

if ((sys.version_info[0] == 3) and (sys.version_info[1] >= 8)) or (
    sys.version_info[0] >= 4
):
    from typing import Literal, TypedDict  # noqa: F401
elif (sys.version_info[0] == 3) and (sys.version_info[1] in [7]):
    from typing_extensions import Literal, TypedDict  # noqa: F401
else:
    raise RuntimeError(
        "Unsupported Python version, {0}".format(sys.version_info)
    )



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""Static analysis subcommand implementation."""

import logging
import pathlib
import sys

import click

from build_harness._utility import (
    CommandArgs,
    command_path,
    report_console_error,
    run_command,
)
from build_harness.project import acquire_source_dir

from .code_style import FormattingError, _apply_formatting
from .state import CommandState, ExitState

log = logging.getLogger(__name__)


FLAKE8_FORMAT_CMD: CommandArgs = ["flake8", "."]
MYPY_FORMAT_CMD: CommandArgs = ["mypy", "--show-error-codes", "."]

VALID_ANALYSIS_OPTIONS = [
    "all",
    "flake8",
    "mypy",
]


class StaticAnalysisError(Exception):
    """Problem occurred during static analysis."""


class Flake8Error(StaticAnalysisError):
    """Problem occurred with flake8 analysis."""


class MypyError(StaticAnalysisError):
    """Problem occurred with mypy analysis."""


def _run_flake8(source_path: pathlib.Path, venv_path: pathlib.Path) -> None:
    """
    Run flake8 analysis on the specified source directory.

    Args:
        source_path: Path to project source directory.
        venv_path: Path to Python virtual environment.

    Raises:
        StaticAnalysisError: If flake8 analysis fails.
    """
    click.echo("flake8 analysis...")
    FLAKE8_FORMAT_CMD[0] = command_path(venv_path, FLAKE8_FORMAT_CMD)
    FLAKE8_FORMAT_CMD[-1] = str(source_path)
    result = run_command(FLAKE8_FORMAT_CMD)

    if result.returncode != 0:
        raise Flake8Error("flake8 analysis failed")
    click.echo(click.style("flake8 analysis completed", fg="green"))


def _run_mypy(source_path: pathlib.Path, venv_path: pathlib.Path) -> None:
    """
    Run mypy analysis on the specified source directory.

    Args:
        source_path: Path to project source directory.
        venv_path: Path to Python virtual environment.

    Raises:
        StaticAnalysisError: If mypy analysis fails.
    """
    click.echo("mypy analysis...")
    MYPY_FORMAT_CMD[0] = command_path(venv_path, MYPY_FORMAT_CMD)
    MYPY_FORMAT_CMD[-1] = str(source_path)
    result = run_command(MYPY_FORMAT_CMD)

    if result.returncode != 0:
        raise MypyError("mypy analysis failed")
    click.echo(click.style("mypy analysis completed", fg="green"))


def _apply_analysis(
    analysis_type: str, project_path: pathlib.Path, venv_path: pathlib.Path
) -> None:
    """
    Apply the specified analysis.

    Args:
        project_path: Path to project root directory.
        venv_path: Path to Python virtual environment.
    """
    source_dir = acquire_source_dir(project_path)

    analysis_methods = {
        "flake8": _run_flake8,
        "mypy": _run_mypy,
    }
    if analysis_type != "all":
        analysis_methods[analysis_type](project_path / source_dir, venv_path)
    else:
        for this_method in [y for x, y in analysis_methods.items()]:
            this_method(project_path / source_dir, venv_path)


@click.command()
@click.option(
    "--analysis",
    default="all",
    help="Analysis type to apply (default all)",
    type=click.Choice(VALID_ANALYSIS_OPTIONS, case_sensitive=True),
)
@click.pass_context
def static_analysis(ctx: click.Context, analysis: str) -> None:
    """
    Run suite of Python static analysis.

    The suite currently comprises "flake8" and "mypy". Formatting is run before
    the static analysis to eliminate formatting complaints from static analysis
    tools.
    """
    try:
        ctx.ensure_object(dict)
        command_state: CommandState = ctx.obj["command_state"]

        _apply_formatting(command_state.venv_path)
        _apply_analysis(
            analysis, command_state.project_path, command_state.venv_path
        )
    except FormattingError as e:
        report_console_error(str(e))
        sys.exit(ExitState.FORMATTING_FAILED.value)
    except Flake8Error as e:
        report_console_error(str(e))
        sys.exit(ExitState.FLAKE8_FAILED.value)
    except MypyError as e:
        report_console_error(str(e))
        sys.exit(ExitState.MYPY_FAILED.value)
    except Exception:
        message = "Unexpected error. Check log for details."
        log.exception(message)
        report_console_error(message)
        sys.exit(ExitState.UNKNOWN_ERROR.value)



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#


"""Acceptance tests subcommand implementation."""

import logging
import pathlib

from build_harness._utility import command_path, run_command

log = logging.getLogger(__name__)


def _generate_snippets(venv_path: pathlib.Path) -> None:
    behave_command = [
        "behave",
        "--snippets",
        "--tags",
        "@notimplemented",
        "features",
    ]
    behave_command[0] = command_path(venv_path, behave_command)
    # ignore errors from the snippets command because it seems to always exit
    # dirty
    run_command(behave_command)



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""
Organise tool abstractions for tools consumed by ``build_harness``.

Exports:

*  ``Pep426Patterns`` class
*  ``Pep440Patterns`` class
*  ``PytestCommand`` class
*  ``pep503_normalize`` function
"""

from .pep426 import Pep426Patterns  # noqa :F401
from .pep440 import Pep440Patterns  # noqa :F401
from .pep503 import pep503_normalize  # noqa :F401
from .pytest import (  # noqa :F401
    VALID_COVERAGE_REPORTS,
    CommandOptionsTestTool,
    PytestCommand,
)



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""PEP-503 related functions."""

import re


def pep503_normalize(name: str) -> str:
    """
    Normalize a package name per PEP-503.

    Assumes the name is compliant with PEP-426.

    https://www.python.org/dev/peps/pep-0503/#normalized-names
    https://www.python.org/dev/peps/pep-0426/

    Args:
        name: PEP-426 compliant package name.

    Returns:
        PEP-503 normalized name.
    """
    return re.sub(r"[-_.]+", "-", name).lower()



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""CLI command processing for ``build-harness`` command."""

import logging
import pathlib
import sys

import click
from click_logging_config import logging_parameters

from build_harness._version import __version__, acquire_version

from ._declarations import DEFAULT_BUILDHARNESS_LOGCONFIG, DEFAULT_PROJECT_PATH
from .analysis import static_analysis
from .bdd import bdd_acceptance_command
from .bootstrap import bootstrap
from .code_style import formatting
from .dependencies import install
from .publishing import publish
from .state import CommandState
from .unit_tests import unit_test
from .wheel import package

log = logging.getLogger(__name__)


@click.group()
@click.pass_context
@click.version_option(version=acquire_version())
@logging_parameters(DEFAULT_BUILDHARNESS_LOGCONFIG)
def main(
    ctx: click.Context,
) -> None:
    """Build harness group."""
    ctx.ensure_object(dict)
    ctx.obj["command_state"] = CommandState(
        project_path=pathlib.Path(DEFAULT_PROJECT_PATH),
        venv_path=pathlib.Path(sys.argv[0]).parent.absolute(),
    )
    log.info(f"build harness version, {__version__}")


main.add_command(bdd_acceptance_command, name="acceptance")
main.add_command(bootstrap, name="bootstrap")
main.add_command(formatting, name="formatting")
main.add_command(install, name="install")
main.add_command(package, name="package")
main.add_command(publish, name="publish")
main.add_command(static_analysis, name="static-analysis")
main.add_command(unit_test, name="unit-test")



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import pathlib
import tempfile

from build_harness.commands.dependencies import _check_virtual_environment


class TestCheckVirtualEnvironment:
    def test_clean(self):
        with tempfile.TemporaryDirectory() as directory:
            mock_project_dir = pathlib.Path(directory)

            venv_path = _check_virtual_environment(mock_project_dir)

            assert venv_path.is_dir()



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""Common constants for command definitions."""

import enum
import pathlib

from click_logging_config import LoggingConfiguration

DEFAULT_PROJECT_PATH = "."

DEFAULT_CONSOLE_LOGGING_ENABLED = False
DEFAULT_FILE_LOGGING_ENABLED = True
DEFAULT_FILE_ROTATION_BACKUPS = 10
DEFAULT_FILE_ROTATION_SIZE_MB = 1
DEFAULT_LOG_FILE = pathlib.Path("build_harness.log")
DEFAULT_LOG_LEVEL = "warning"

DEFAULT_BUILDHARNESS_LOGCONFIG = LoggingConfiguration.parse_obj(
    {
        "enable_console_logging": DEFAULT_CONSOLE_LOGGING_ENABLED,
        "enable_file_logging": DEFAULT_FILE_LOGGING_ENABLED,
        "file_logging": {
            "file_rotation_size_megabytes": DEFAULT_FILE_ROTATION_SIZE_MB,
            "log_file_path": DEFAULT_LOG_FILE,
            "max_rotation_backup_files": DEFAULT_FILE_ROTATION_BACKUPS,
        },
        "log_level": DEFAULT_LOG_LEVEL,
    }
)


@enum.unique
class PublishOptions(enum.Enum):
    """Enumeration of artifact publish options."""

    yes = enum.auto()
    no = enum.auto()
    dryrun = enum.auto()
    test = enum.auto()



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import pathlib
import subprocess

import git
import pytest

from build_harness.commands import main
from build_harness.commands.wheel import _apply_release_id, _build_flit_package
from tests.ci.support.click_runner import click_runner  # noqa: F401


@pytest.fixture()
def mock_run(mock_sysargv, mocker):
    mock_pytest_result = mocker.create_autospec(subprocess.CompletedProcess)
    mock_pytest_result.returncode = 0
    mock_pytest_result.stdout = ""
    this_run = mocker.patch(
        "build_harness.commands.wheel.run_command",
        return_value=mock_pytest_result,
    )

    return this_run


class TestApplyReleaseId:
    def test_clean(self, mocker):
        mock_open = mocker.patch(
            "build_harness.commands.wheel.pathlib.Path.open"
        )
        mock_repo = mocker.patch(
            "build_harness.commands.wheel.git.Repo", autospec=True
        )

        mock_path = pathlib.Path("some/test_wheel_clean/path")
        mock_release = "1.2.3+abc3.14"

        _apply_release_id(mock_release, mock_path)

        mock_handle = mock_open.return_value.__enter__.return_value
        mock_handle.write.assert_called_once_with(mock_release)

        mock_repo.return_value.index.add.assert_called_once_with(
            [str(mock_path)]
        )
        mock_repo.return_value.index.commit.assert_called_once_with(
            str(mock_path)
        )

    def test_not_git_repo(self, mocker):
        mock_open = mocker.patch(
            "build_harness.commands.wheel.pathlib.Path.open"
        )
        mock_repo = mocker.patch(
            "build_harness.commands.wheel.git.Repo",
            autospec=True,
            side_effect=git.InvalidGitRepositoryError(),
        )

        mock_path = pathlib.Path("some/test_wheel_not_git_repo/path")
        mock_release = "1.2.3+abc3.14"

        _apply_release_id(mock_release, mock_path)

        mock_handle = mock_open.return_value.__enter__.return_value
        mock_handle.write.assert_called_once_with(mock_release)

        mock_repo.return_value.index.add.assert_not_called()
        mock_repo.return_value.index.commit.assert_not_called()


class TestPackageBuild:
    def test_default(self, click_runner, mocker, mock_run):
        mock_apply_release = mocker.patch(
            "build_harness.commands.wheel._apply_release_id"
        )
        result = click_runner.invoke(
            main,
            [
                "package",
            ],
        )

        assert result.exit_code == 0
        mock_run.assert_has_calls(
            [
                mocker.call(
                    ["/some/conftest/path/flit", "build"],
                ),
            ]
        )
        mock_apply_release.assert_not_called()

    def test_release_id_set(self, click_runner, mocker, mock_run):
        mock_apply_release = mocker.patch(
            "build_harness.commands.wheel._apply_release_id"
        )

        expected_id = "3.14.159"

        result = click_runner.invoke(
            main,
            [
                "package",
                "--release-id",
                expected_id,
            ],
        )

        assert result.exit_code == 0
        mock_run.assert_has_calls(
            [
                mocker.call(
                    ["/some/conftest/path/flit", "build"],
                ),
            ]
        )
        mock_apply_release.assert_called_once_with(expected_id, mocker.ANY)

    def test_empty_release_id(self, click_runner, mocker, mock_run):
        mock_apply_release = mocker.patch(
            "build_harness.commands.wheel._apply_release_id"
        )
        result = click_runner.invoke(
            main,
            [
                "package",
                "--release-id",
                "",
            ],
        )

        assert result.exit_code == 0
        mock_run.assert_has_calls(
            [
                mocker.call(
                    ["/some/conftest/path/flit", "build"],
                ),
            ]
        )
        mock_apply_release.assert_not_called()



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""Initialize a new project from a template."""

import logging
import pathlib
import sys

import click
from cookiecutter.main import cookiecutter  # type: ignore

from build_harness._utility import report_console_error

from .state import ExitState

log = logging.getLogger(__name__)

VALID_CI_CHOICES = ["gitlabci"]
VALID_PACKAGER_CHOICES = ["flit"]


@click.command()
@click.pass_context
@click.argument(
    "project_name",
    type=str,
)
@click.option(
    "--ci",
    default=VALID_CI_CHOICES[0],
    help="Select CI pipeline file(s) to be generated.",
    type=click.Choice(VALID_CI_CHOICES),
)
@click.option(
    "--default-branch",
    default="main",
    help="Select CI pipeline file(s) to be generated.",
    type=str,
)
@click.option(
    "--disable-dependencies",
    default=False,
    help="Disable build harness dependency installation (development only). If "
    "disabled it is assumed the user has pre-installed the necessary "
    "dependencies in the correct location.",
    is_flag=True,
)
@click.option(
    "--force",
    default=False,
    help="Write over any existing files.",
    is_flag=True,
)
@click.option(
    "--packager",
    default=VALID_PACKAGER_CHOICES[0],
    help="Select Python packaging utility to be used.",
    type=click.Choice(VALID_PACKAGER_CHOICES),
)
def bootstrap(
    ctx: click.Context,
    ci: str,
    default_branch: str,
    disable_dependencies: bool,
    force: bool,
    packager: str,
    project_name: str,
) -> None:
    """
    Initialize a new Python project from a template.

    The new project includes initial unit tests, coverage analysis, working CI
    pipeline, packaging and publishing to pypi.org.
    """
    try:
        template_dir = pathlib.Path(__file__).parent / "../templates"

        context_configuration = {
            "default_branch": default_branch,
            "packaging_provider": packager,
            "pipeline_provider": ci,
            "project_name": project_name,
            # Regular users should not need this option, but it is needed for
            # integration tests to acquire the correct build_harness version for
            # testing.
            "disable_bh_dependencies": disable_dependencies,
        }
        cookiecutter(
            str(template_dir),
            extra_context=context_configuration,
            no_input=True,
            replay=False,
            # NOTE: require overwrite_if_exists==True because project directory
            #       is populated with .venv before cookiecutter is run.
            #       Otherwise cookiecutter will fail.
            overwrite_if_exists=True,
            output_dir=str(pathlib.Path.cwd()),
        )
    except Exception as e:
        message = "Unexpected error. Check log for details."
        log.exception(message)
        log.error(str(e))
        report_console_error(message)
        sys.exit(ExitState.UNKNOWN_ERROR.value)



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#


"""Publish flow command entry point."""

import logging
import pathlib
import sys
import typing

import click
from click_logging_config import LoggingConfiguration, logging_parameters

from build_harness._utility import report_console_error
from build_harness._version import __version__, acquire_version
from build_harness.tools.git import (
    DEFAULT_DEFAULT_BRANCH_NAME,
    GitNotFoundError,
    GitRepoError,
    get_tag_data,
    validate_git,
)

from ._declarations import (
    DEFAULT_CONSOLE_LOGGING_ENABLED,
    DEFAULT_FILE_LOGGING_ENABLED,
    DEFAULT_PROJECT_PATH,
    PublishOptions,
)
from ._release_id import InvalidReleaseId, validate_release_id
from .state import CommandState, ExitState

log = logging.getLogger(__name__)

DEFAULT_PUBLISHFLOW_LOGCONFIG = LoggingConfiguration.parse_obj(
    {
        "enable_file_logging": DEFAULT_FILE_LOGGING_ENABLED,
        "enable_console_logging": DEFAULT_CONSOLE_LOGGING_ENABLED,
        "file_logging": {"log_file_path": pathlib.Path("publish_flow.log")},
        "log_level": "debug",
    }
)

VALID_PR_IDENTIFIERS = {
    # Gitlab-CI
    "merge_request_event",
}


@click.command()
@click.version_option(version=acquire_version())
@click.option(
    "--default-branch",
    default=DEFAULT_DEFAULT_BRANCH_NAME,
    help="Git default/main branch name.",
    show_default=True,
    type=str,
)
@click.option(
    "--disable-pr-publish",
    default=None,
    help="""Disable pull/merge request pipeline publishing artifacts.

For Gitlab-CI TEXT must be the value of CI_PIPELINE_SOURCE variable.

[default: enabled - PR/MR pipeline behaves the same as feature branch
pipeline]""",
    type=str,
)
@click.option(
    "--git",
    default=None,
    help="Git executable. [default: users PATH]",
    type=str,
)
@click.option(
    "--project",
    default=DEFAULT_PROJECT_PATH,
    help="Git project directory.",
    show_default=True,
    type=str,
)
@click.option(
    "--publish-nonrelease",
    default=False,
    help="Publish on untagged non-releases.",
    is_flag=True,
    show_default=True,
)
@click.option(
    "--publish-prerelease",
    default=False,
    help="Publish on a semantic version pre-release tag.",
    is_flag=True,
    show_default=True,
)
@logging_parameters(DEFAULT_PUBLISHFLOW_LOGCONFIG)
def publish_flow_main(
    default_branch: str,
    disable_pr_publish: typing.Optional[str],
    git: typing.Optional[str],
    project: str,
    publish_nonrelease: bool,
    publish_prerelease: bool,
) -> None:
    """
    Determine whether to publish from git tag and commit history.

    By default, if HEAD commit is tagged with a semantic version release then
    publish is activated. Otherwise publishing is deactivated.

    The utility emits the ``--publish`` or ``--no-publish`` option to be
    consumed by the ``build-harness publish`` subcommand.
    """
    try:
        state = CommandState(
            project_path=pathlib.Path(project),
            venv_path=pathlib.Path(sys.argv[0]).parent.absolute(),
        )
        log.info(f"build harness version, {__version__}")

        validate_git(git)

        tag_data = get_tag_data(state.project_path, default_branch)
        tag_version = validate_release_id(tag_data.tag)
        publish_result = PublishOptions.test.name
        if (
            (tag_data.offset and publish_nonrelease)
            or (
                tag_version.is_prerelease
                and (not tag_data.offset)
                and (publish_prerelease or publish_nonrelease)
            )
            or ((not tag_version.is_prerelease) and (not tag_data.offset))
        ) and (not tag_version.local):
            publish_result = PublishOptions.yes.name
        elif (tag_version.local and tag_version.local.startswith("dryrun")) or (
            disable_pr_publish and (disable_pr_publish in VALID_PR_IDENTIFIERS)
        ):
            publish_result = PublishOptions.dryrun.name

        click.echo(publish_result, err=False, nl=False)
    except InvalidReleaseId as e:
        report_console_error(str(e))
        sys.exit(ExitState.BAD_VERSION.value)
    except GitRepoError as e:
        report_console_error(str(e))
        sys.exit(ExitState.BAD_REPO.value)
    except GitNotFoundError as e:
        report_console_error(str(e))
        click.echo("Git must be installed and configured to use this utility.")
        click.echo("In Linux, something like this is necessary:")
        click.echo("    sudo apt install -y git || sudo yum install git")
        click.echo('    git config user.name "Your Name"')
        click.echo('    git config user.email "you@example.com"')
        sys.exit(ExitState.BAD_GIT_EXE.value)
    except Exception as e:
        message = "Unexpected error. Check log for details."
        log.exception(str(e))
        report_console_error(message)
        sys.exit(ExitState.UNKNOWN_ERROR.value)



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import typing

import pytest

from build_harness.commands import ExitState, main
from tests.ci.support.click_runner import click_runner  # noqa: F401


@pytest.fixture
def run_bootstrap(click_runner, mocker):
    def _apply(command_arguments: typing.List[str], this_runner):
        mock_cookiecutter = mocker.patch(
            "build_harness.commands.bootstrap.cookiecutter"
        )
        arguments = ["bootstrap"] + command_arguments
        result = this_runner.invoke(main, arguments)

        return result, mock_cookiecutter

    return _apply


class TestBootstrap:
    def test_default(self, click_runner, run_bootstrap, mocker):
        result, mock_cookiecutter = run_bootstrap(
            ["my_new_project"], click_runner
        )

        assert result.exit_code == 0
        mock_cookiecutter.assert_called_once_with(
            mocker.ANY,
            extra_context={
                "default_branch": "main",
                "disable_bh_dependencies": False,
                "packaging_provider": "flit",
                "pipeline_provider": "gitlabci",
                "project_name": "my_new_project",
            },
            no_input=True,
            replay=False,
            overwrite_if_exists=True,
            output_dir=mocker.ANY,
        )

    def test_ci(self, click_runner, run_bootstrap, mocker):
        result, mock_cookiecutter = run_bootstrap(
            ["my_new_project", "--ci", "gitlabci"], click_runner
        )

        assert result.exit_code == 0
        mock_cookiecutter.assert_called_once_with(
            mocker.ANY,
            extra_context={
                "default_branch": "main",
                "disable_bh_dependencies": False,
                "packaging_provider": "flit",
                "pipeline_provider": "gitlabci",
                "project_name": "my_new_project",
            },
            no_input=True,
            replay=False,
            overwrite_if_exists=True,
            output_dir=mocker.ANY,
        )

    def test_packager(self, click_runner, run_bootstrap, mocker):
        result, mock_cookiecutter = run_bootstrap(
            ["my_new_project", "--packager", "flit"], click_runner
        )

        assert result.exit_code == 0
        mock_cookiecutter.assert_called_once_with(
            mocker.ANY,
            extra_context={
                "default_branch": "main",
                "disable_bh_dependencies": False,
                "packaging_provider": "flit",
                "pipeline_provider": "gitlabci",
                "project_name": "my_new_project",
            },
            no_input=True,
            replay=False,
            overwrite_if_exists=True,
            output_dir=mocker.ANY,
        )

    def test_default_branch(self, click_runner, run_bootstrap, mocker):
        result, mock_cookiecutter = run_bootstrap(
            ["my_new_project", "--default-branch", "fancy"], click_runner
        )

        assert result.exit_code == 0
        mock_cookiecutter.assert_called_once_with(
            mocker.ANY,
            extra_context={
                "default_branch": "fancy",
                "disable_bh_dependencies": False,
                "packaging_provider": "flit",
                "pipeline_provider": "gitlabci",
                "project_name": "my_new_project",
            },
            no_input=True,
            replay=False,
            overwrite_if_exists=True,
            output_dir=mocker.ANY,
        )

    def test_force(self, click_runner, run_bootstrap, mocker):
        result, mock_cookiecutter = run_bootstrap(
            ["my_new_project", "--force"], click_runner
        )

        assert result.exit_code == 0
        mock_cookiecutter.assert_called_once_with(
            mocker.ANY,
            extra_context={
                "default_branch": "main",
                "disable_bh_dependencies": False,
                "packaging_provider": "flit",
                "pipeline_provider": "gitlabci",
                "project_name": "my_new_project",
            },
            no_input=True,
            replay=False,
            overwrite_if_exists=True,
            output_dir=mocker.ANY,
        )

    def test_bad_ci_choice_error(self, click_runner, mocker):
        mocker.patch("build_harness.commands.bootstrap.cookiecutter")
        result = click_runner.invoke(
            main, ["bootstrap", "new_project", "--ci", "bad_choice"]
        )

        assert result.exit_code == 2

    def test_bad_packager_choice_error(self, click_runner, mocker):
        mocker.patch("build_harness.commands.bootstrap.cookiecutter")
        result = click_runner.invoke(
            main, ["bootstrap", "new_project", "--packager", "bad_choice"]
        )

        assert result.exit_code == 2

    def test_unknown_error(self, click_runner, mocker):
        mocker.patch(
            "build_harness.commands.bootstrap.cookiecutter",
            side_effect=RuntimeError(),
        )
        result = click_runner.invoke(main, ["bootstrap", "new_project"])

        assert result.exit_code == ExitState.UNKNOWN_ERROR.value



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import pathlib
import typing

import pytest

from tests.ci.support.project_context import ProjectData, run_cookiecutter


@pytest.fixture(scope="module")
def flit_gitlabci_project():
    project_name = "test-flit-gitlabci"
    project_slug = project_name.replace(" ", "_").replace("-", "_")
    context_configuration = {
        "packaging_provider": "flit",
        "pipeline_provider": "gitlabci",
        "project_name": project_name,
        "disable_bh_dependencies": True,
    }
    with run_cookiecutter(context_configuration, project_name) as (
        this_context,
        captured_stdio,
    ):
        temp_path = this_context.temp_path

        expected_files: typing.List[pathlib.Path] = [
            temp_path / project_slug / "pyproject.toml",
            temp_path / project_slug / ".gitlab-ci.yml",
        ]
        file_contents = dict()
        for this_file in expected_files:
            if this_file.is_file():
                with this_file.open("r") as f:
                    file_contents[this_file.name] = f.read()
            else:
                file_contents[this_file.name] = None

    result = ProjectData(
        captured_io=captured_stdio,
        file_contents=file_contents,
        project_name=project_name,
    )

    return result


class TestFlitGitlabciCookiecutter:
    def test_flit(self, flit_gitlabci_project):
        if not flit_gitlabci_project.file_contents["pyproject.toml"]:
            flit_gitlabci_project.captured_io.err.seek(0)
            flit_gitlabci_project.captured_io.out.seek(0)

            print(flit_gitlabci_project.captured_io.out.read())
            print(flit_gitlabci_project.captured_io.err.read())
            pytest.fail("pyproject.toml file not found")

        assert (
            'requires = ["flit_core'
            in flit_gitlabci_project.file_contents["pyproject.toml"]
        )

    def test_gitlabci(self, flit_gitlabci_project):
        if not flit_gitlabci_project.file_contents[".gitlab-ci.yml"]:
            flit_gitlabci_project.captured_io.err.seek(0)
            flit_gitlabci_project.captured_io.out.seek(0)

            print(flit_gitlabci_project.captured_io.out.read())
            print(flit_gitlabci_project.captured_io.err.read())
            pytest.fail(".gitlab-ci.yml file not found")



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""PEP-426 related regular expression patterns."""


class Pep426Patterns:
    """PEP-426 regular expression patterns."""

    UNNORMALIZED_PACKAGE_NAME = r"(?P<package_name>[A-Za-z0-9\-_.]+)"
